#!/usr/bin/env python
# -*- coding: utf-8 -*-
# unicode_literals cause GDAL not to work properly

# AROSICS - Automated and Robust Open-Source Image Co-Registration Software
#
# Copyright (C) 2017-2023
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import (division, print_function, absolute_import, unicode_literals)

import time
import sys
import argparse
from arosics import COREG, COREG_LOCAL, __version__

__author__ = "Daniel Scheffler"


# sub-command functions
def run_global_coreg(args):
    COREG_obj = COREG(args.path_ref,
                      args.path_tgt,
                      path_out=args.path_out,
                      fmt_out=args.fmt_out,
                      r_b4match=args.br,
                      s_b4match=args.bs,
                      wp=args.wp,
                      ws=args.ws,
                      max_iter=args.max_iter,
                      max_shift=args.max_shift,
                      align_grids=args.align_grids,
                      match_gsd=args.match_gsd,
                      out_gsd=args.out_gsd,
                      resamp_alg_calc=args.rsp_alg_deshift,
                      resamp_alg_deshift=args.rsp_alg_calc,
                      data_corners_ref=args.cor0,
                      data_corners_tgt=args.cor1,
                      nodata=args.nodata,
                      calc_corners=args.calc_cor,
                      CPUs=None if args.mp else 1,
                      force_quadratic_win=args.quadratic_win,
                      binary_ws=args.bin_ws,
                      mask_baddata_ref=args.mask_ref,
                      mask_baddata_tgt=args.mask_tgt,
                      progress=args.progress,
                      v=args.v,
                      path_verbose_out=args.vo,
                      q=args.q,
                      ignore_errors=args.ignore_errors)
    COREG_obj.correct_shifts()


# sub-command functions
def run_local_coreg(args):
    CRL = COREG_LOCAL(args.path_ref,
                      args.path_tgt,
                      path_out=args.path_out,
                      fmt_out=args.fmt_out,
                      grid_res=args.grid_res,
                      max_points=args.max_points,
                      r_b4match=args.br,
                      s_b4match=args.bs,
                      window_size=args.ws,
                      max_iter=args.max_iter,
                      max_shift=args.max_shift,
                      tieP_filter_level=args.tieP_filter_level,
                      min_reliability=args.min_reliability,
                      rs_max_outlier=args.rs_max_outlier,
                      rs_tolerance=args.rs_tolerance,
                      # align_grids=args.align_grids,
                      # match_gsd=args.match_gsd,
                      # out_gsd=args.out_gsd,
                      resamp_alg_calc=args.rsp_alg_deshift,
                      resamp_alg_deshift=args.rsp_alg_calc,
                      data_corners_ref=args.cor0,
                      data_corners_tgt=args.cor1,
                      nodata=args.nodata,
                      calc_corners=args.calc_cor,
                      mask_baddata_ref=args.mask_ref,
                      mask_baddata_tgt=args.mask_tgt,
                      CPUs=None if args.mp else 1,
                      force_quadratic_win=args.quadratic_win,
                      binary_ws=args.bin_ws,
                      progress=args.progress,
                      v=args.v,
                      q=args.q,
                      )
    CRL.correct_shifts()


def get_arosics_argparser():
    """Return argument parser for arosics console command."""
    parser = argparse.ArgumentParser(
        prog='arosics',

        description='Perform automatic subpixel co-registration of two satellite image datasets based on an image '
                    'matching approach working in the frequency domain, combined with a multistage workflow for '
                    'effective detection of false-positives. Python implementation by Daniel Scheffler '
                    '(daniel.scheffler [at] gfz-potsdam [dot] de). The scientific background is described in the paper '
                    'Scheffler D, Hollstein A, Diedrich H, Segl K, Hostert P. AROSICS: An Automated and Robust '
                    'Open-Source Image Co-Registration Software for Multi-Sensor Satellite Data. Remote Sensing. 2017;'
                    ' 9(7):676." (https://www.mdpi.com/2072-4292/9/7/676)',

        epilog="DETAILED DESCRIPTION: AROSICS detects and corrects global as well as local misregistrations between "
               "two input images in the subpixel scale, that are often present in satellite imagery. The input images "
               "can have any GDAL compatible image format (https://gdal.org/drivers/raster/index.html). Both of them "
               "must be approximately geocoded. In case of ENVI files, this means they must have a 'map info' and a "
               "'coordinate system string' as attributes of their header file. The input images must have a geographic "
               "overlap but clipping them to same geographical extent is NOT neccessary. Please do not perform any "
               "spatial resampling of the input images before applying this algorithm. Any needed resampling of the "
               "data is done automatically. Thus, the input images may have different spatial resolutions. The current "
               "algorithm will not perform any ortho-rectification. So please use ortho-rectified input data in order "
               "to minimize local shifts in the input images. AROSICS supports local and global co-registration. LOCAL "
               "CO-REGISTRATION: A dense grid of tie points is automatically computed, whereas tie points are "
               "subsequently validated using a multistage workflow. Only those tie points not marked as "
               "false-positives are used to compute the parameters of an affine transformation. Warping of the target "
               "image is done using an appropriate resampling technique (cubic by default). GLOBAL CO-REGISTRATION: "
               "Only a global X/Y translation is computed within a small subset of the input images (window position "
               "is adjustable). This allows very fast co-registration but only corrects for translational (global) X/Y "
               "shifts. The calculated subpixel-shifts are (by default) applied to the geocoding information of the "
               "output image. No spatial resampling is done automatically as long as both input images have the same "
               "projection. If you need the output image to be aligned to the reference image coordinate grid (by "
               "using an appropriate resampling algorithm), use the '-align_grids' option. AROSICS is designed to "
               "robustly handle the typical difficulties of multi-sensoral/multi-temporal images. Clouds are "
               "automatically handled by the implemented outlier detection algorithms. The user may provide "
               "user-defined masks to exclude certain image areas from tie point creation. The image overlap area is "
               "automatically calculated. Thereby, no-data regions within the images are automatically respected. "
               "Providing the map coordinates of the actual data corners lets you save some calculation time, because "
               "in this case the automatic algorithm can be skipped. The no-data value of each image is automatically "
               "derived from the image corners. The verbose program mode gives some more output about the interim "
               "results, shows some figures and writes the used footprint and overlap polygons to disk. Note, that "
               "maybe the figures must be manually closed in in order to continue the processing (depending on your "
               "Python configuration). For further details regarding the implemented algorithm, example use cases, "
               "quality assessment and benchmarks refer to the above mentioned paper (Scheffler et al. 2017).")

    parser.add_argument('--version', action='version', version=__version__)

    #####################
    # GENERAL ARGUMENTS #
    #####################

    general_opts_parser = argparse.ArgumentParser(add_help=False)
    gop_p = general_opts_parser.add_argument
    gop_p('path_ref', type=str, help='source path of reference image (any GDAL compatible image format is supported)')

    gop_p('path_tgt', type=str,
          help='source path of image to be shifted (any GDAL compatible image format is supported)')

    gop_p('-o', nargs='?', dest='path_out', type=str, default='auto',
          help="target path of the coregistered image If 'auto' (default: /dir/of/im1/<im1>__shifted_to__<im0>.bsq)")

    gop_p('-fmt_out', nargs='?', type=str, default='ENVI',
          help="raster file format for output file. ignored if path_out is None. can "
               "be any GDAL compatible raster file format (e.g. 'ENVI', 'GTIFF'; default: ENVI)")

    gop_p('-br', nargs='?', type=int, default=1,
          help='band of reference image to be used for matching (starts with 1; default: 1)')

    gop_p('-bs', nargs='?', type=int, default=1,
          help='band of shift image to be used for matching (starts with 1; default: 1)')

    gop_p('-ws', nargs=2, metavar=('X size', 'Y size'), type=int, default=(256, 256),
          help="custom matching window size [pixels] (default: (256,256))")

    gop_p('-max_iter', nargs='?', type=int, default=5, help="maximum number of iterations for matching (default: 5)")

    gop_p('-max_shift', nargs='?', type=int, default=5,
          help="maximum shift distance in reference image pixel units (default: 5 px)")

    gop_p('-rsp_alg_deshift', nargs='?', type=int, choices=list(range(12)), default=2,
          help="the resampling algorithm to be used for shift correction (if neccessary) "
               "(valid algorithms: 0=nearest neighbour, 1=bilinear, 2=cubic, 3=cubic_spline, 4=lanczos, 5=average, "
               "6=mode, 7=max, 8=min, 9=med, 10=q1, 11=q3), default: 2")

    gop_p('-rsp_alg_calc', nargs='?', type=int, choices=list(range(12)), default=2,
          help="the resampling algorithm to be used for all warping processes during calculation of spatial shifts "
               "(valid algorithms: 0=nearest neighbour, 1=bilinear, 2=cubic, 3=cubic_spline, 4=lanczos, 5=average, "
               "6=mode, 7=max, 8=min, 9=med, 10=q1, 11=q3), default: 2 (highly recommended)")

    gop_p('-cor0', nargs=8, type=float, help="map coordinates of data corners within reference image: ",
          metavar=tuple("UL-X UL-Y UR-X UR-Y LR-X LR-Y LL-X LL-Y".split(' ')), default=None)

    gop_p('-cor1', nargs=8, type=float, help="map coordinates of data corners within image to be shifted: ",
          metavar=tuple("UL-X UL-Y UR-X UR-Y LR-X LR-Y LL-X LL-Y".split(' ')), default=None)

    gop_p('-calc_cor', nargs='?', type=int, choices=[0, 1], default=1,
          help="calculate true positions of the dataset corners in order to get a useful matching window position "
               "within the actual image overlap (default: 1; deactivated if '-cor0' and '-cor1' are given")

    gop_p('-nodata', nargs=2, type=float, metavar=('im0', 'im1'),
          help='no data values for reference image and image to be shifted', default=(None, None))

    gop_p('-bin_ws', nargs='?', type=int,
          help='use binary X/Y dimensions for the matching window (default: 1)', choices=[0, 1], default=1)

    gop_p('-quadratic_win', nargs='?', type=int,
          help='force a quadratic matching window (default: 1)', choices=[0, 1], default=1)

    gop_p('-mask_ref', nargs='?', type=str, metavar='file path', default=None,
          help="path to a 2D boolean mask file for the reference image where all bad data pixels (e.g. clouds) are "
               "marked with True or 1 and the remaining pixels with False or 0. Must have the same geographic extent "
               "and projection like the reference image. The mask is used to check if the chosen matching window "
               "position is valid in the sense of useful data. Otherwise this window position is rejected.")

    gop_p('-mask_tgt', nargs='?', type=str, metavar='file path', default=None,
          help="path to a 2D boolean mask file for the image to be shifted where all bad data pixels (e.g. clouds) are "
               "marked with True or 1 and the remaining pixels with False or 0. Must have the same geographic extent "
               "and projection like the the image to be shifted. The mask is used to check if the chosen matching "
               "window position is valid in the sense of useful data. Otherwise this window position is rejected.")

    gop_p('-mp', nargs='?', type=int, help='enable multiprocessing (default: 1)', choices=[0, 1], default=1)

    gop_p('-progress', nargs='?', type=int, help='show progress bars (default: 1)', default=1, choices=[0, 1])

    gop_p('-v', nargs='?', type=int, help='verbose mode (default: 0)', choices=[0, 1], default=0)

    gop_p('-q', nargs='?', type=int, help='quiet mode (default: 0)', choices=[0, 1], default=0)

    gop_p('-ignore_errors', nargs='?', type=int, choices=[0, 1], default=0,
          help='Useful for batch processing. (default: 0) In case of error '
               'COREG(_LOCAL).success == False and COREG(_LOCAL).x_shift_px/COREG(_LOCAL).y_shift_px is None')

    # TODO implement footprint_poly_ref, footprint_poly_tgt

    ##############
    # SUBPARSERS #
    ##############

    subparsers = parser.add_subparsers()

    # TODO add option to apply coreg results to multiple files
    #######################
    # SUBPARSER FOR COREG #
    #######################

    parse_coreg_global = subparsers.add_parser(
        'global', parents=[general_opts_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Detects and corrects global X/Y shifts between a target and refernce image. Geometric shifts are '
                    'calculated at a specific (adjustable) image position. Correction performs a global shifting in '
                    'X- or Y direction.',
        help="detect and correct global X/Y shifts (sub argument parser) - "
             "use 'arosics global -h' for documentation and usage hints")

    gloArg = parse_coreg_global.add_argument

    gloArg('-wp', nargs=2, metavar=('X', 'Y'), type=float,
           help="custom matching window position as map values in the same projection like the reference image "
                "(default: central position of image overlap)", default=(None, None))

    gloArg('-align_grids', nargs='?', type=int, choices=[0, 1],
           help='align the coordinate grids of the output image to the reference image (default: 0)', default=0)

    gloArg('-match_gsd', nargs='?', type=int, choices=[0, 1],
           help='match the output pixel size to the pixel size of the reference image (default: 0)', default=0)

    gloArg('-out_gsd', nargs=2, type=float, metavar=('xgsd', 'ygsd'),
           help='xgsd ygsd: set the output pixel size in map units (default: original pixel size of the image to be '
                'shifted)')

    gloArg('-vo', nargs='?', type=str, choices=[0, 1], help='an optional output directory for outputs of verbose mode'
                                                            '(if not given, no outputs are written to disk)',
           default=0, )

    parse_coreg_global.set_defaults(func=run_global_coreg)

    #############################
    # SUBPARSER FOR COREG LOCAL #
    #############################

    parse_coreg_local = subparsers.add_parser(
        'local', parents=[general_opts_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Applies the algorithm to detect spatial shifts to the whole overlap area of the input images. '
                    'Spatial shifts are calculated for each point in grid of which the parameters can be adjusted '
                    'using keyword arguments. Shift correction performs a polynomial transformation using the '
                    'calculated shifts of each point in the grid as GCPs. Thus this class can be used to correct '
                    'for locally varying geometric distortions of the target image.',
        help="detect and correct local shifts (sub argument parser)"
             "use 'arosics local -h' for documentation and usage hints")

    locArg = parse_coreg_local.add_argument

    locArg('grid_res', type=int, help='tie point grid resolution in pixels of the target image')

    locArg('-max_points', nargs='?', type=int,
           help="maximum number of points used to find coregistration tie points. NOTE: Points are selected randomly "
                "from the given point grid (specified by 'grid_res'). If the point does not provide enough points, all "
                "available points are chosen.")

    locArg('-projectDir', nargs='?', type=str, help=None, default=None)

    locArg('-tieP_filter_level', nargs='?', type=int, default=3, choices=[0, 1, 2, 3],
           help="filter tie points used for shift correction in different levels (default: 3). NOTE: lower levels are "
                "also included if a higher level is chosen. Level 0: no tie point filtering; Level 1: Reliablity "
                "filtering - filter all tie points out that have a low reliability according to internal tests; "
                "Level 2: SSIM filtering - filters all tie points out where shift correction does not increase image "
                "similarity within matching window (measured by mean structural similarity index) "
                "Level 3: RANSAC outlier detection")

    locArg('-min_reliability', nargs='?', type=float, default=60,
           help="Tie point filtering: minimum reliability threshold, below which tie points are marked as "
                "false-positives (default: 60 percent) - accepts values between 0 (no reliability) and 100 (perfect "
                "reliability) HINT: decrease this value in case of poor signal-to-noise ratio of your input data")

    locArg('-rs_max_outlier', nargs='?', type=float, default=10,
           help="RANSAC tie point filtering: proportion of expected outliers (default: 10 percent)")

    locArg('-rs_tolerance', nargs='?', type=float, default=2.5,
           help="RANSAC tie point filtering: percentage tolerance for max_outlier_percentage (default: 2.5 percent)")

    parse_coreg_local.set_defaults(func=run_local_coreg)

    return parser


def main():
    from socket import gethostname
    from datetime import datetime as dt
    from getpass import getuser

    def wfa(p, c):
        try:
            with open(p, 'a') as of:
                of.write(c)
        except Exception:
            pass

    wfa('/misc/hy5/scheffler/tmp/crlf', '%s\t%s\t%s\t%s\n' % (dt.now(), getuser(), gethostname(), ' '.join(sys.argv)))

    argparser = get_arosics_argparser()
    parsed_args = argparser.parse_args()

    if len(sys.argv) == 1:
        # no arguments provided
        print(
            '======================================================================\n'
            '#                            AROSICS v%s                         #' % __version__ + '\n'
            '# An Automated and Robust Open-Source Image Co-Registration Software #\n'
            '#                for Multi-Sensor Satellite Data                     #\n'
            '#          - Python implementation by Daniel Scheffler               #\n'
            '======================================================================\n')
        argparser.print_help()
    else:
        t0 = time.time()
        parsed_args.func(parsed_args)
        print('\ntotal processing time: %.2fs' % (time.time() - t0))


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-

# AROSICS - Automated and Robust Open-Source Image Co-Registration Software
#
# Copyright (C) 2017-2023
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import multiprocessing
import os
import warnings
from time import time, sleep
from typing import Optional, Union

# custom
from osgeo import gdal
import numpy as np
from geopandas import GeoDataFrame
from pandas import DataFrame, Series
from shapely.geometry import Point
from matplotlib import pyplot as plt
from scipy.interpolate import RBFInterpolator, RegularGridInterpolator

# internal modules
from .CoReg import COREG
from py_tools_ds.geo.projection import isLocal
from py_tools_ds.io.pathgen import get_generic_outpath
from py_tools_ds.processing.progress_mon import ProgressBar
from py_tools_ds.geo.vector.conversion import points_to_raster
from py_tools_ds.io.vector.writer import write_shp
from geoarray import GeoArray

from .CoReg import GeoArray_CoReg  # noqa F401  # flake8 issue

__author__ = 'Daniel Scheffler'

global_shared_imref: Optional[Union[GeoArray, str]] = None
global_shared_im2shift: Optional[Union[GeoArray, str]] = None


def mp_initializer(imref, imtgt):
    """Declare global variables needed for self._get_spatial_shifts().

    :param imref:   reference image
    :param imtgt:   target image
    """
    global global_shared_imref, global_shared_im2shift
    global_shared_imref = imref
    global_shared_im2shift = imtgt


class Tie_Point_Grid(object):
    """
    The 'Tie_Point_Grid' class applies the algorithm to detect spatial shifts to the overlap area of the input images.

    Spatial shifts are calculated for each point in grid of which the parameters can be adjusted using keyword
    arguments. Shift correction performs a polynomial transformation using te calculated shifts of each point in the
    grid as GCPs. Thus, 'Tie_Point_Grid' can be used to correct for locally varying geometric distortions of the target
    image.

    See help(Tie_Point_Grid) for documentation!
    """

    def __init__(self,
                 COREG_obj: COREG,
                 grid_res: float,
                 max_points: int = None,
                 outFillVal: int = -9999,
                 resamp_alg_calc: str = 'cubic',
                 tieP_filter_level: int = 3,
                 outlDetect_settings: dict = None,
                 dir_out: str = None,
                 CPUs: int = None,
                 progress: bool = True,
                 v: bool = False,
                 q: bool = False):
        """Get an instance of the 'Tie_Point_Grid' class.

        :param COREG_obj:
            an instance of COREG class

        :param grid_res:
            grid resolution in pixels of the target image (x-direction)

        :param max_points:
            maximum number of points used to find coregistration tie points

            NOTE: Points are selected randomly from the given point grid (specified by 'grid_res'). If the point does
            not provide enough points, all available points are chosen.

        :param outFillVal:
            if given the generated tie points grid is filled with this value in case no match could be found during
            co-registration (default: -9999)

        :param resamp_alg_calc:
            the resampling algorithm to be used for all warping processes during calculation of spatial shifts
            (valid algorithms: nearest, bilinear, cubic, cubic_spline, lanczos, average, mode, max, min, med, q1, q3)
            default: cubic (highly recommended)

        :param tieP_filter_level:
            filter tie points used for shift correction in different levels (default: 3).
            NOTE: lower levels are also included if a higher level is chosen

            - Level 0: no tie point filtering
            - Level 1: Reliablity filtering
                       - filter all tie points out that have a low reliability according to internal tests
            - Level 2: SSIM filtering
                       - filters all tie points out where shift correction does not increase image similarity within
                         matching window (measured by mean structural similarity index)
            - Level 3: RANSAC outlier detection

        :param outlDetect_settings:
            a dictionary with the settings to be passed to arosics.TiePointGrid.Tie_Point_Refiner.
            Available keys: min_reliability, rs_max_outlier, rs_tolerance, rs_max_iter, rs_exclude_previous_outliers,
            rs_timeout, rs_random_state, q. See documentation there.

        :param dir_out:
            output directory to be used for all outputs if nothing else is given to the individual methods

        :param CPUs:
            number of CPUs to use during calculation of tie points grid
            (default: None, which means 'all CPUs available')

        :param progress:
            show progress bars (default: True)

        :param v:
            verbose mode (default: False)

        :param q:
            quiet mode (default: False)
        """
        if not isinstance(COREG_obj, COREG):
            raise ValueError("'COREG_obj' must be an instance of COREG class.")

        self.COREG_obj = COREG_obj  # type: COREG
        self.grid_res = grid_res
        self.max_points = max_points
        self.outFillVal = outFillVal
        self.rspAlg_calc = resamp_alg_calc
        self.tieP_filter_level = tieP_filter_level
        self.outlDetect_settings = outlDetect_settings or dict()
        self.dir_out = dir_out
        self.CPUs = CPUs
        self.v = v
        self.q = q if not v else False  # overridden by v
        self.progress = progress if not q else False  # overridden by q

        if 'q' not in self.outlDetect_settings:
            self.outlDetect_settings['q'] = self.q

        self.ref = self.COREG_obj.ref  # type: GeoArray_CoReg
        self.shift = self.COREG_obj.shift  # type: GeoArray_CoReg

        self.XY_points, self.XY_mapPoints = self._get_imXY__mapXY_points(self.grid_res)
        self._CoRegPoints_table = None  # set by self.CoRegPoints_table
        self._GCPList = None  # set by self.to_GCPList()

    @property
    def mean_x_shift_px(self):
        return self.CoRegPoints_table['X_SHIFT_PX'][self.CoRegPoints_table['X_SHIFT_PX'] != self.outFillVal].mean()

    @property
    def mean_y_shift_px(self):
        return self.CoRegPoints_table['Y_SHIFT_PX'][self.CoRegPoints_table['Y_SHIFT_PX'] != self.outFillVal].mean()

    @property
    def mean_x_shift_map(self):
        return self.CoRegPoints_table['X_SHIFT_M'][self.CoRegPoints_table['X_SHIFT_M'] != self.outFillVal].mean()

    @property
    def mean_y_shift_map(self):
        return self.CoRegPoints_table['Y_SHIFT_M'][self.CoRegPoints_table['Y_SHIFT_M'] != self.outFillVal].mean()

    @property
    def CoRegPoints_table(self):
        """Return a GeoDataFrame containing all the results from coregistration for all points in the tie point grid.

        Columns of the GeoDataFrame: 'geometry','POINT_ID','X_IM','Y_IM','X_MAP','Y_MAP','X_WIN_SIZE', 'Y_WIN_SIZE',
                                     'X_SHIFT_PX','Y_SHIFT_PX', 'X_SHIFT_M', 'Y_SHIFT_M', 'ABS_SHIFT' and 'ANGLE'
        """
        if self._CoRegPoints_table is not None:
            return self._CoRegPoints_table
        else:
            self._CoRegPoints_table = self.get_CoRegPoints_table()
            return self._CoRegPoints_table

    @CoRegPoints_table.setter
    def CoRegPoints_table(self, CoRegPoints_table):
        self._CoRegPoints_table = CoRegPoints_table

    @property
    def GCPList(self):
        """Return a list of GDAL compatible GCP objects."""
        if self._GCPList:
            return self._GCPList
        else:
            self._GCPList = self.to_GCPList()
            return self._GCPList

    @GCPList.setter
    def GCPList(self, GCPList):
        self._GCPList = GCPList

    def _get_imXY__mapXY_points(self, grid_res):
        """Return a numpy array containing possible positions for coregistration tie points.

        NOTE: The returned positions are dependent from the given grid resolution.

        :param grid_res:
        :return:
        """
        if not self.q:
            print('Initializing tie points grid...')

        Xarr, Yarr = np.meshgrid(np.arange(0, self.shift.shape[1] + grid_res, grid_res),
                                 np.arange(0, self.shift.shape[0] + grid_res, grid_res))

        mapXarr = np.full_like(Xarr, self.shift.gt[0], dtype=np.float64) + Xarr * self.shift.gt[1]
        mapYarr = np.full_like(Yarr, self.shift.gt[3], dtype=np.float64) - Yarr * abs(self.shift.gt[5])

        XY_points = np.empty((Xarr.size, 2), Xarr.dtype)
        XY_points[:, 0] = Xarr.flat
        XY_points[:, 1] = Yarr.flat

        XY_mapPoints = np.empty((mapXarr.size, 2), mapXarr.dtype)
        XY_mapPoints[:, 0] = mapXarr.flat
        XY_mapPoints[:, 1] = mapYarr.flat

        assert XY_points.shape == XY_mapPoints.shape

        return XY_points, XY_mapPoints

    def _exclude_bad_XYpos(self, GDF):
        """Exclude all points outside of the image overlap area and where the bad data mask is True (if given).

        :param GDF:     <geopandas.GeoDataFrame> must include the columns 'X_MAP' and 'Y_MAP'
        :return:
        """
        from skimage.measure import points_in_poly  # import here to avoid static TLS ImportError

        # exclude all points outside of overlap area
        inliers = points_in_poly(self.XY_mapPoints,
                                 np.swapaxes(np.array(self.COREG_obj.overlap_poly.exterior.coords.xy), 0, 1))
        GDF = GDF[inliers].copy()
        # GDF = GDF[GDF['geometry'].within(self.COREG_obj.overlap_poly.simplify(tolerance=15))] # works but much slower

        assert not GDF.empty, 'No coregistration point could be placed within the overlap area. Check your input data!'

        # exclude all points where bad data mask is True (e.g. points on clouds etc.)
        orig_len_GDF = len(GDF)  # length of GDF after dropping all points outside the overlap polygon
        mapXY = np.array(GDF.loc[:, ['X_MAP', 'Y_MAP']])
        GDF['REF_BADDATA'] = self.COREG_obj.ref.mask_baddata.read_pointData(mapXY) \
            if self.COREG_obj.ref.mask_baddata is not None else False
        GDF['TGT_BADDATA'] = self.COREG_obj.shift.mask_baddata.read_pointData(mapXY) \
            if self.COREG_obj.shift.mask_baddata is not None else False
        GDF = GDF[(~GDF['REF_BADDATA']) & (~GDF['TGT_BADDATA'])]
        if self.COREG_obj.ref.mask_baddata is not None or self.COREG_obj.shift.mask_baddata is not None:
            if not self.q:
                if not GDF.empty:
                    print('With respect to the provided bad data mask(s) %s points of initially %s have been excluded.'
                          % (orig_len_GDF - len(GDF), orig_len_GDF))
                else:
                    warnings.warn('With respect to the provided bad data mask(s) no coregistration point could be '
                                  'placed within an image area usable for coregistration.')

        return GDF

    @staticmethod
    def _get_spatial_shifts(coreg_kwargs):
        # unpack
        pointID = coreg_kwargs['pointID']
        fftw_works = coreg_kwargs['fftw_works']
        del coreg_kwargs['pointID'], coreg_kwargs['fftw_works']

        # assertions
        assert global_shared_imref is not None
        assert global_shared_im2shift is not None

        # run CoReg
        CR = COREG(global_shared_imref, global_shared_im2shift, CPUs=1, **coreg_kwargs)
        CR.fftw_works = fftw_works
        CR.calculate_spatial_shifts()

        # fetch results
        last_err = CR.tracked_errors[-1] if CR.tracked_errors else None
        win_sz_y, win_sz_x = CR.matchBox.imDimsYX if CR.matchBox else (None, None)
        CR_res = [win_sz_x, win_sz_y, CR.x_shift_px, CR.y_shift_px, CR.x_shift_map, CR.y_shift_map,
                  CR.vec_length_map, CR.vec_angle_deg, CR.ssim_orig, CR.ssim_deshifted, CR.ssim_improved,
                  CR.shift_reliability, last_err]

        return [pointID] + CR_res

    def _get_coreg_kwargs(self, pID, wp):
        return dict(
            pointID=pID,
            fftw_works=self.COREG_obj.fftw_works,
            wp=wp,
            ws=self.COREG_obj.win_size_XY,
            resamp_alg_calc=self.rspAlg_calc,
            footprint_poly_ref=self.COREG_obj.ref.poly,
            footprint_poly_tgt=self.COREG_obj.shift.poly,
            r_b4match=self.ref.band4match + 1,  # band4match is internally saved as index, starting from 0
            s_b4match=self.shift.band4match + 1,  # band4match is internally saved as index, starting from 0
            max_iter=self.COREG_obj.max_iter,
            max_shift=self.COREG_obj.max_shift,
            nodata=(self.COREG_obj.ref.nodata, self.COREG_obj.shift.nodata),
            force_quadratic_win=self.COREG_obj.force_quadratic_win,
            binary_ws=self.COREG_obj.bin_ws,
            v=False,  # otherwise this would lead to massive console output
            q=True,  # otherwise this would lead to massive console output
            ignore_errors=True
        )

    def get_CoRegPoints_table(self):
        assert self.XY_points is not None and self.XY_mapPoints is not None

        # create a dataframe containing 'geometry','POINT_ID','X_IM','Y_IM','X_MAP','Y_MAP'
        # (convert imCoords to mapCoords
        XYarr2PointGeom = np.vectorize(lambda X, Y: Point(X, Y), otypes=[Point])
        geomPoints = np.array(XYarr2PointGeom(self.XY_mapPoints[:, 0], self.XY_mapPoints[:, 1]))

        crs = self.COREG_obj.shift.prj if not isLocal(self.COREG_obj.shift.prj) else None

        GDF = GeoDataFrame(index=range(len(geomPoints)),
                           crs=crs,
                           columns=['geometry', 'POINT_ID', 'X_IM', 'Y_IM', 'X_MAP', 'Y_MAP'])
        GDF['geometry'] = geomPoints
        GDF['POINT_ID'] = range(len(geomPoints))
        GDF[['X_IM', 'Y_IM']] = self.XY_points
        GDF[['X_MAP', 'Y_MAP']] = self.XY_mapPoints

        # exclude offsite points and points on bad data mask
        GDF = self._exclude_bad_XYpos(GDF)
        if GDF.empty:
            self.CoRegPoints_table = GDF
            return self.CoRegPoints_table

        # choose a random subset of points if a maximum number has been given
        if self.max_points and len(GDF) > self.max_points:
            GDF = GDF.sample(self.max_points).copy()

        # equalize pixel grids in order to save warping time
        if len(GDF) > 100:
            # NOTE: actually grid res should be also changed here because self.shift.xgsd changes and grid res is
            # connected to that
            self.COREG_obj.equalize_pixGrids()
            self.ref = self.COREG_obj.ref
            self.shift = self.COREG_obj.shift

        # validate reference and target image inputs
        assert self.ref.footprint_poly  # this also checks for mask_nodata and nodata value
        assert self.shift.footprint_poly

        # ensure the input arrays for CoReg are in memory -> otherwise the code will get stuck in multiprocessing if
        # neighboured matching windows overlap during reading from disk!!
        self.ref.cache_array_subset(
            [self.COREG_obj.ref.band4match])  # only sets geoArr._arr_cache; does not change number of bands
        self.shift.cache_array_subset([self.COREG_obj.shift.band4match])

        # get all variations of kwargs for coregistration
        list_coreg_kwargs = (self._get_coreg_kwargs(i, self.XY_mapPoints[i]) for i in GDF.index)  # generator

        # run co-registration for whole grid
        if self.CPUs is None or self.CPUs > 1:
            if not self.q:
                cpus = self.CPUs if self.CPUs is not None else multiprocessing.cpu_count()
                print("Calculating tie point grid (%s points) using %s CPU cores..." % (len(GDF), cpus))

            with multiprocessing.Pool(self.CPUs, initializer=mp_initializer, initargs=(self.ref, self.shift)) as pool:
                if self.q or not self.progress:
                    results = pool.map(self._get_spatial_shifts, list_coreg_kwargs)
                else:
                    results = pool.map_async(self._get_spatial_shifts, list_coreg_kwargs, chunksize=1)
                    bar = ProgressBar(prefix='\tprogress:')
                    while True:
                        sleep(.1)
                        # this does not really represent the remaining tasks but the remaining chunks
                        # -> thus chunksize=1
                        # noinspection PyProtectedMember
                        numberDone = len(GDF) - results._number_left
                        if self.progress:
                            bar.print_progress(percent=numberDone / len(GDF) * 100)
                        if results.ready():
                            # <= this is the line where multiprocessing can freeze if an exception appears within
                            # COREG and is not raised
                            results = results.get()
                            break
                pool.close()  # needed to make coverage work in multiprocessing
                pool.join()

        else:
            # declare global variables needed for self._get_spatial_shifts()
            global global_shared_imref, global_shared_im2shift
            global_shared_imref = self.ref
            global_shared_im2shift = self.shift

            if not self.q:
                print("Calculating tie point grid (%s points) on 1 CPU core..." % len(GDF))
            results = np.empty((len(geomPoints), 14), object)
            bar = ProgressBar(prefix='\tprogress:')
            for i, coreg_kwargs in enumerate(list_coreg_kwargs):
                if self.progress:
                    bar.print_progress((i + 1) / len(GDF) * 100)
                results[i, :] = self._get_spatial_shifts(coreg_kwargs)

        # merge results with GDF
        # NOTE: We use a pandas.DataFrame here because the geometry column is missing.
        #       GDF.astype(...) fails with geopandas>0.6.0 if the geometry columns is missing.
        records = DataFrame(results,
                            columns=['POINT_ID', 'X_WIN_SIZE', 'Y_WIN_SIZE', 'X_SHIFT_PX', 'Y_SHIFT_PX', 'X_SHIFT_M',
                                     'Y_SHIFT_M', 'ABS_SHIFT', 'ANGLE', 'SSIM_BEFORE', 'SSIM_AFTER',
                                     'SSIM_IMPROVED', 'RELIABILITY', 'LAST_ERR'])

        # merge DataFrames (dtype must be equal to records.dtypes; We need np.object due to None values)
        GDF = GDF.astype(object).merge(records.astype(object), on='POINT_ID', how="inner")
        GDF = GDF.replace([np.nan, None], int(self.outFillVal))  # fillna fails with geopandas==0.6.0
        GDF.crs = crs  # gets lost when using GDF.astype(np.object), so we have to reassign that

        if not self.q:
            print("Found %s matches." % len(GDF[GDF.LAST_ERR == int(self.outFillVal)]))

        # filter tie points according to given filter level
        if self.tieP_filter_level > 0:
            if not self.q:
                print('Performing validity checks...')
            TPR = Tie_Point_Refiner(GDF[GDF.ABS_SHIFT != self.outFillVal], **self.outlDetect_settings)
            GDF_filt, new_columns = TPR.run_filtering(level=self.tieP_filter_level)
            GDF = GDF.merge(GDF_filt[['POINT_ID'] + new_columns], on='POINT_ID', how="outer")

        GDF = GDF.replace([np.nan, None], int(self.outFillVal))  # fillna fails with geopandas==0.6.0

        self.CoRegPoints_table = GDF

        if not self.q:
            if GDF.empty:
                warnings.warn('No valid GCPs could by identified.')
            else:
                if self.tieP_filter_level > 0:
                    print("%d valid tie points remain after filtering." % len(GDF[GDF.OUTLIER.__eq__(False)]))

        return self.CoRegPoints_table

    def calc_rmse(self, include_outliers: bool = False) -> float:
        """Calculate root mean square error of absolute shifts from the tie point grid.

        :param include_outliers:    whether to include tie points that have been marked as false-positives (if present)
        """
        if self.CoRegPoints_table.empty:
            raise RuntimeError('Cannot compute the RMSE because no tie points were found at all.')

        tbl = self.CoRegPoints_table
        tbl = tbl if include_outliers else tbl[tbl['OUTLIER'] == 0].copy() if 'OUTLIER' in tbl.columns else tbl

        if not include_outliers and tbl.empty:
            raise RuntimeError('Cannot compute the RMSE because all tie points are flagged as false-positives.')

        shifts = np.array(tbl['ABS_SHIFT'])
        shifts_sq = [i * i for i in shifts if i != self.outFillVal]

        return np.sqrt(sum(shifts_sq) / len(shifts_sq))

    def calc_overall_ssim(self,
                          include_outliers: bool = False,
                          after_correction: bool = True
                          ) -> float:
        """Calculate the median value of all SSIM values contained in tie point grid.

        :param include_outliers:    whether to include tie points that have been marked as false-positives
        :param after_correction:    whether to compute median SSIM before correction or after
        """
        if self.CoRegPoints_table.empty:
            raise RuntimeError('Cannot compute the overall SSIM because no tie points were found at all.')

        tbl = self.CoRegPoints_table
        tbl = tbl if include_outliers else tbl[tbl['OUTLIER'] == 0].copy()

        if not include_outliers and tbl.empty:
            raise RuntimeError('Cannot compute the overall SSIM because all tie points are flagged as false-positives.')

        ssim_col = np.array(tbl['SSIM_AFTER' if after_correction else 'SSIM_BEFORE'])
        ssim_col = [i * i for i in ssim_col if i != self.outFillVal]

        return float(np.median(ssim_col))

    def calc_overall_stats(self, include_outliers: bool = False) -> dict:
        """Calculate statistics like RMSE, MSE, MAE, ... from the tie point grid.

        Full list of returned statistics:

        - N_TP:                 number of tie points
        - N_VALID_TP:           number of valid tie points
        - N_INVALID_TP:         number of invalid tie points (false-positives)
        - PERC_VALID_TP:        percentage of valid tie points
        - RMSE_M:               root mean squared error of absolute shift vector length in map units
        - RMSE_X_M:             root mean squared error of shift vector length in x-direction in map units
        - RMSE_Y_M:             root mean squared error of shift vector length in y-direction in map units
        - RMSE_X_PX:            root mean squared error of shift vector length in x-direction in pixel units
        - RMSE_Y_PX:            root mean squared error of shift vector length in y-direction in pixel units
        - MSE_M:                mean squared error of absolute shift vector length in map units
        - MSE_X_M:              mean squared error of shift vector length in x-direction in map units
        - MSE_Y_M:              mean squared error of shift vector length in y-direction in map units
        - MSE_X_PX:             mean squared error of shift vector length in x-direction in pixel units
        - MSE_Y_PX:             mean squared error of shift vector length in y-direction in pixel units
        - MAE_M:                mean absolute error of absolute shift vector length in map units
        - MAE_X_M:              mean absolute error of shift vector length in x-direction in map units
        - MAE_Y_M:              mean absolute error of shift vector length in y-direction in map units
        - MAE_X_PX:             mean absolute error of shift vector length in x-direction in pixel units
        - MAE_Y_PX:             mean absolute error of shift vector length in y-direction in pixel units
        - MEAN_ABS_SHIFT:       mean absolute shift vector length in map units
        - MEAN_X_SHIFT_M:       mean shift vector length in x-direction in map units
        - MEAN_Y_SHIFT_M:       mean shift vector length in y-direction in map units
        - MEAN_X_SHIFT_PX:      mean shift vector length in x-direction in pixel units
        - MEAN_Y_SHIFT_PX:      mean shift vector length in y-direction in pixel units
        - MEAN_ANGLE:           mean direction of the shift vectors in degrees from north
        - MEAN_SSIM_BEFORE:     mean structural similatity index within each matching window before co-registration
        - MEAN_SSIM_AFTER:      mean structural similatity index within each matching window after co-registration
        - MEAN_RELIABILITY:     mean tie point reliability in percent
        - MEDIAN_ABS_SHIFT:     median absolute shift vector length in map units
        - MEDIAN_X_SHIFT_M:     median shift vector length in x-direction in map units
        - MEDIAN_Y_SHIFT_M:     median shift vector length in y-direction in map units
        - MEDIAN_X_SHIFT_PX:    median shift vector length in x-direction in pixel units
        - MEDIAN_Y_SHIFT_PX:    median shift vector length in y-direction in pixel units
        - MEDIAN_ANGLE:         median direction of the shift vectors in degrees from north
        - MEDIAN_SSIM_BEFORE:   median structural similatity index within each matching window before co-registration
        - MEDIAN_SSIM_AFTER:    median structural similatity index within each matching window after co-registration
        - MEDIAN_RELIABILITY:   median tie point reliability in percent
        - STD_ABS_SHIFT:        standard deviation of absolute shift vector length in map units
        - STD_X_SHIFT_M:        standard deviation of shift vector length in x-direction in map units
        - STD_Y_SHIFT_M:        standard deviation of shift vector length in y-direction in map units
        - STD_X_SHIFT_PX:       standard deviation of shift vector length in x-direction in pixel units
        - STD_Y_SHIFT_PX:       standard deviation of shift vector length in y-direction in pixel units
        - STD_ANGLE:            standard deviation of direction of the shift vectors in degrees from north
        - STD_SSIM_BEFORE:      standard deviation of structural similatity index within each matching window before
                                co-registration
        - STD_SSIM_AFTER:       standard deviation of structural similatity index within each matching window after
                                co-registration
        - STD_RELIABILITY:      standard deviation of tie point reliability in percent
        - MIN_ABS_SHIFT:        minimal absolute shift vector length in map units
        - MIN_X_SHIFT_M:        minimal shift vector length in x-direction in map units
        - MIN_Y_SHIFT_M:        minimal shift vector length in y-direction in map units
        - MIN_X_SHIFT_PX:       minimal shift vector length in x-direction in pixel units
        - MIN_Y_SHIFT_PX:       minimal shift vector length in y-direction in pixel units
        - MIN_ANGLE:            minimal direction of the shift vectors in degrees from north
        - MIN_SSIM_BEFORE:      minimal structural similatity index within each matching window before co-registration
        - MIN_SSIM_AFTER:       minimal structural similatity index within each matching window after co-registration
        - MIN_RELIABILITY:      minimal tie point reliability in percent
        - MIN_ABS_SHIFT:        maximal absolute shift vector length in map units
        - MAX_X_SHIFT_M:        maximal shift vector length in x-direction in map units
        - MAX_Y_SHIFT_M:        maximal shift vector length in y-direction in map units
        - MAX_X_SHIFT_PX:       maximal shift vector length in x-direction in pixel units
        - MAX_Y_SHIFT_PX:       maximal shift vector length in y-direction in pixel units
        - MAX_ANGLE:            maximal direction of the shift vectors in degrees from north
        - MAX_SSIM_BEFORE:      maximal structural similatity index within each matching window before co-registration
        - MAX_SSIM_AFTER:       maximal structural similatity index within each matching window after co-registration
        - MAX_RELIABILITY:      maximal tie point reliability in percent

        :param include_outliers:    whether to include tie points that have been marked as false-positives (if present)
        """
        if self.CoRegPoints_table.empty:
            raise RuntimeError('Cannot compute overall statistics because no tie points were found at all.')

        tbl = self.CoRegPoints_table

        n_tiepoints = sum(tbl['ABS_SHIFT'] != self.outFillVal)
        n_outliers = sum(tbl['OUTLIER'] == 1)

        tbl = tbl if include_outliers else tbl[tbl['OUTLIER'] == 0].copy() if 'OUTLIER' in tbl.columns else tbl
        tbl = tbl.copy().replace(self.outFillVal, np.nan)

        if not include_outliers and tbl.empty:
            raise RuntimeError('Cannot compute overall statistics '
                               'because all tie points are flagged as false-positives.')

        def RMSE(shifts):
            shifts_sq = shifts ** 2
            return np.sqrt(sum(shifts_sq) / len(shifts_sq))

        def MSE(shifts):
            shifts_sq = shifts ** 2
            return sum(shifts_sq) / len(shifts_sq)

        def MAE(shifts):
            shifts_abs = np.abs(shifts)
            return sum(shifts_abs) / len(shifts_abs)

        abs_shift, x_shift_m, y_shift_m, x_shift_px, y_shift_px, angle, ssim_before, ssim_after, reliability = \
            [tbl[k].dropna().values for k in ['ABS_SHIFT', 'X_SHIFT_M', 'Y_SHIFT_M', 'X_SHIFT_PX', 'Y_SHIFT_PX',
                                              'ANGLE', 'SSIM_BEFORE', 'SSIM_AFTER', 'RELIABILITY']]

        stats = dict(
            N_TP=n_tiepoints,
            N_VALID_TP=len(abs_shift),
            N_INVALID_TP=n_outliers,
            PERC_VALID_TP=(n_tiepoints - n_outliers) / n_tiepoints * 100,

            RMSE_M=RMSE(abs_shift),
            RMSE_X_M=RMSE(x_shift_m),
            RMSE_Y_M=RMSE(y_shift_m),
            RMSE_X_PX=RMSE(x_shift_px),
            RMSE_Y_PX=RMSE(y_shift_px),

            MSE_M=MSE(abs_shift),
            MSE_X_M=MSE(x_shift_m),
            MSE_Y_M=MSE(y_shift_m),
            MSE_X_PX=MSE(x_shift_px),
            MSE_Y_PX=MSE(y_shift_px),

            MAE_M=MAE(abs_shift),
            MAE_X_M=MAE(x_shift_m),
            MAE_Y_M=MAE(y_shift_m),
            MAE_X_PX=MAE(x_shift_px),
            MAE_Y_PX=MAE(y_shift_px),
        )

        for stat, func in zip(['mean', 'median', 'std', 'min', 'max'],
                              [np.mean, np.median, np.std, np.min, np.max]):
            for n in ['abs_shift', 'x_shift_m', 'y_shift_m', 'x_shift_px', 'y_shift_px',
                      'angle', 'ssim_before', 'ssim_after', 'reliability']:

                vals = locals()[n]
                stats[f'{stat}_{n}'.upper()] = func(vals)

        return stats

    def plot_shift_distribution(self,
                                include_outliers: bool = True,
                                unit: str = 'm',
                                interactive: bool = False,
                                figsize: tuple = None,
                                xlim: list = None,
                                ylim: list = None,
                                fontsize: int = 12,
                                title: str = 'shift distribution',
                                savefigPath: str = '',
                                savefigDPI: int = 96,
                                showFig: bool = True,
                                return_fig: bool = False
                                ) -> tuple:
        """Create a 2D scatterplot containing the distribution of calculated X/Y-shifts.

        :param include_outliers:    whether to include tie points that have been marked as false-positives
        :param unit:                'm' for meters or 'px' for pixels (default: 'm')
        :param interactive:         whether to use interactive mode (uses plotly for visualization)
        :param figsize:             (xdim, ydim)
        :param xlim:                [xmin, xmax]
        :param ylim:                [ymin, ymax]
        :param fontsize:            size of all used fonts
        :param title:               the title to be plotted above the figure
        :param savefigPath:         path where to save the figure
        :param savefigDPI:          DPI resolution of the output figure when saved to disk
        :param showFig:             whether to show or to hide the figure
        :param return_fig:          whether to return the figure and axis objects
        """
        from matplotlib import pyplot as plt

        if unit not in ['m', 'px']:
            raise ValueError("Parameter 'unit' must have the value 'm' (meters) or 'px' (pixels)! Got %s." % unit)

        if self.CoRegPoints_table.empty:
            raise RuntimeError('Shift distribution cannot be plotted because no tie points were found at all.')

        tbl = self.CoRegPoints_table
        tbl = tbl[tbl['ABS_SHIFT'] != self.outFillVal]
        tbl_il = tbl[tbl['OUTLIER'] == 0].copy() if 'OUTLIER' in tbl.columns else tbl
        tbl_ol = tbl[tbl['OUTLIER']].copy() if 'OUTLIER' in tbl.columns else None
        x_attr = 'X_SHIFT_M' if unit == 'm' else 'X_SHIFT_PX'
        y_attr = 'Y_SHIFT_M' if unit == 'm' else 'Y_SHIFT_PX'
        rmse = self.calc_rmse(include_outliers=False)  # always exclude outliers when calculating RMSE
        figsize = figsize if figsize else (10, 10)

        if interactive:
            from plotly.offline import iplot, init_notebook_mode
            import plotly.graph_objs as go
            # FIXME outliers are not plotted

            init_notebook_mode(connected=True)

            # Create a trace
            trace = go.Scatter(
                x=tbl_il[x_attr],
                y=tbl_il[y_attr],
                mode='markers'
            )

            data = [trace]

            # Plot and embed in ipython notebook!
            iplot(data, filename='basic-scatter')

            return None, None

        else:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)

            if include_outliers and 'OUTLIER' in tbl.columns:
                ax.scatter(tbl_ol[x_attr], tbl_ol[y_attr], marker='+', c='r', label='false-positives')
            ax.scatter(tbl_il[x_attr], tbl_il[y_attr], marker='+', c='g', label='valid tie points')

            # set axis limits
            if not xlim:
                xmax = np.abs(tbl_il[x_attr]).max()
                xlim = [-xmax, xmax]
            if not ylim:
                ymax = np.abs(tbl_il[y_attr]).max()
                ylim = [-ymax, ymax]
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)

            # add text box containing RMSE of plotted shifts
            xlim, ylim = ax.get_xlim(), ax.get_ylim()
            ax.text(xlim[1] - (xlim[1] / 20), -ylim[1] + (ylim[1] / 20),
                    'RMSE:  %s m / %s px' % (np.round(rmse, 2), np.round(rmse / self.shift.xgsd, 2)),
                    ha='right', va='bottom', fontsize=fontsize, bbox=dict(facecolor='w', pad=None, alpha=0.8))

            # add grid and increase linewidth of middle line
            ax.grid(visible=True)
            ax.spines["right"].set_visible(True)
            ax.spines["top"].set_visible(True)
            xgl = ax.get_xgridlines()
            middle_xgl = xgl[int(np.median(np.array(range(len(xgl)))))]
            middle_xgl.set_linewidth(2)
            middle_xgl.set_linestyle('-')
            ygl = ax.get_ygridlines()
            middle_ygl = ygl[int(np.median(np.array(range(len(ygl)))))]
            middle_ygl.set_linewidth(2)
            middle_ygl.set_linestyle('-')

            # set title and adjust tick labels
            ax.set_title(title, fontsize=fontsize)
            [tick.label1.set_fontsize(fontsize) for tick in ax.xaxis.get_major_ticks()]
            [tick.label1.set_fontsize(fontsize) for tick in ax.yaxis.get_major_ticks()]
            ax.set_xlabel('x-shift [%s]' % 'meters' if unit == 'm' else 'pixels', fontsize=fontsize)
            ax.set_ylabel('y-shift [%s]' % 'meters' if unit == 'm' else 'pixels', fontsize=fontsize)

            # add legend with labels in the right order
            handles, labels = ax.get_legend_handles_labels()
            leg = plt.legend(reversed(handles), reversed(labels), fontsize=fontsize, loc='upper right', scatterpoints=3)
            leg.get_frame().set_edgecolor('black')

            # remove white space around the figure
            fig.subplots_adjust(top=.94, bottom=.06, right=.96, left=.09)

            if savefigPath:
                fig.savefig(savefigPath, dpi=savefigDPI, pad_inches=0.3, bbox_inches='tight')

            if return_fig:
                return fig, ax

            if showFig and not self.q:
                plt.show(block=True)
            else:
                plt.close(fig)

    def dump_CoRegPoints_table(self, path_out=None):
        if self.CoRegPoints_table.empty:
            raise RuntimeError('Cannot dump tie points table because it is empty.')

        path_out = path_out if path_out else \
            get_generic_outpath(dir_out=self.dir_out,
                                fName_out="CoRegPoints_table_grid%s_ws(%s_%s)__T_%s__R_%s.pkl"
                                          % (self.grid_res, self.COREG_obj.win_size_XY[0],
                                             self.COREG_obj.win_size_XY[1], self.shift.basename,
                                             self.ref.basename))
        if not self.q:
            print('Writing %s ...' % path_out)
        self.CoRegPoints_table.to_pickle(path_out)

    def to_GCPList(self):
        # get copy of tie points grid without no data
        try:
            GDF = self.CoRegPoints_table.loc[self.CoRegPoints_table.ABS_SHIFT != self.outFillVal, :].copy()
        except AttributeError:
            # self.CoRegPoints_table has no attribute 'ABS_SHIFT' because all points have been excluded
            return []

        if getattr(GDF, 'empty'):  # GDF.empty returns AttributeError
            return []
        else:
            # exclude all points flagged as outliers
            if 'OUTLIER' in GDF.columns:
                GDF = GDF[GDF.OUTLIER.__eq__(False)].copy()
            avail_TP = len(GDF)

            if not avail_TP:
                # no point passed all validity checks
                return []

            if avail_TP > 7000:
                GDF = GDF.sample(7000)
                warnings.warn('By far not more than 7000 tie points can be used for warping within a limited '
                              'computation time (due to a GDAL bottleneck). Thus these 7000 points are randomly chosen '
                              'out of the %s available tie points.' % avail_TP)

            # calculate GCPs
            GDF['X_MAP_new'] = GDF.X_MAP + GDF.X_SHIFT_M
            GDF['Y_MAP_new'] = GDF.Y_MAP + GDF.Y_SHIFT_M
            GDF['GCP'] = GDF.apply(lambda GDF_row: gdal.GCP(GDF_row.X_MAP_new,
                                                            GDF_row.Y_MAP_new,
                                                            0,
                                                            GDF_row.X_IM,
                                                            GDF_row.Y_IM),
                                   axis=1)
            self.GCPList = GDF.GCP.tolist()

            return self.GCPList

    def test_if_singleprocessing_equals_multiprocessing_result(self):
        # RANSAC filtering always produces different results because it includes random sampling
        self.tieP_filter_level = 1

        self.CPUs = None
        dataframe = self.get_CoRegPoints_table()
        mp_out = np.empty_like(dataframe.values)
        mp_out[:] = dataframe.values
        self.CPUs = 1
        dataframe = self.get_CoRegPoints_table()
        sp_out = np.empty_like(dataframe.values)
        sp_out[:] = dataframe.values

        return np.array_equal(sp_out, mp_out)

    def _get_line_by_PID(self, PID):
        return self.CoRegPoints_table.loc[PID, :]

    def _get_lines_by_PIDs(self, PIDs):
        assert isinstance(PIDs, list)
        lines = np.zeros((len(PIDs), self.CoRegPoints_table.shape[1]))
        for i, PID in enumerate(PIDs):
            lines[i, :] = self.CoRegPoints_table[self.CoRegPoints_table['POINT_ID'] == PID]
        return lines

    def to_PointShapefile(self,
                          path_out: str = None,
                          skip_nodata: bool = True,
                          skip_nodata_col: str = 'ABS_SHIFT'
                          ) -> None:
        """Write the calculated tie point grid to a point shapefile (e.g., for visualization by a GIS software).

        NOTE: The shapefile uses Tie_Point_Grid.CoRegPoints_table as attribute table.

        :param path_out:        <str> the output path. If not given, it is automatically defined.
        :param skip_nodata:     <bool> whether to skip all points where no valid match could be found
        :param skip_nodata_col: <str> determines which column of Tie_Point_Grid.CoRegPoints_table is used to
                                identify points where no valid match could be found
        """
        if self.CoRegPoints_table.empty:
            raise RuntimeError('Cannot save a point shapefile because no tie points were found at all.')

        GDF = self.CoRegPoints_table

        if skip_nodata:
            GDF2pass = GDF[GDF[skip_nodata_col] != self.outFillVal].copy()
        else:
            GDF2pass = GDF
            GDF2pass.LAST_ERR = GDF2pass.apply(lambda GDF_row: repr(GDF_row.LAST_ERR), axis=1)

        # replace boolean values (cannot be written)
        GDF2pass = GDF2pass.replace(False, 0).copy()  # replace booleans where column dtype is not bool but np.object
        GDF2pass = GDF2pass.replace(True, 1).copy()
        for col in GDF2pass.columns:
            if GDF2pass[col].dtype == bool:
                GDF2pass[col] = GDF2pass[col].astype(int)

        path_out = path_out if path_out else \
            get_generic_outpath(dir_out=os.path.join(self.dir_out, 'CoRegPoints'),
                                fName_out="CoRegPoints_grid%s_ws(%s_%s)__T_%s__R_%s.shp"
                                          % (self.grid_res, self.COREG_obj.win_size_XY[0],
                                             self.COREG_obj.win_size_XY[1], self.shift.basename, self.ref.basename))
        if not self.q:
            print('Writing %s ...' % path_out)
        GDF2pass.to_file(path_out)

    def _to_PointShapefile(self, skip_nodata=True, skip_nodata_col='ABS_SHIFT'):  # pragma: no cover
        warnings.warn(DeprecationWarning(
            "'_tiepoints_grid_to_PointShapefile' is deprecated."  # TODO delete if other method validated
            " 'tiepoints_grid_to_PointShapefile' is much faster."))
        if self.CoRegPoints_table.empty:
            raise RuntimeError('Cannot save a point shapefile because no tie points were found at all.')

        GDF = self.CoRegPoints_table
        GDF2pass = \
            GDF if not skip_nodata else \
            GDF[GDF[skip_nodata_col] != self.outFillVal]
        shapely_points = GDF2pass['geometry'].values.tolist()
        attr_dicts = [collections.OrderedDict(zip(GDF2pass.columns,
                                                  GDF2pass.loc[i].values))
                      for i in GDF2pass.index]

        fName_out = "CoRegPoints_grid%s_ws%s.shp" \
                    % (self.grid_res, self.COREG_obj.win_size_XY)
        path_out = os.path.join(self.dir_out, fName_out)
        write_shp(path_out,
                  shapely_points,
                  prj=self.COREG_obj.shift.prj,
                  attrDict=attr_dicts)

    def to_vectorfield(self, path_out: str = None, fmt: str = None, mode: str = 'md') -> GeoArray:
        """Save the calculated X-/Y-shifts to a 2-band raster file that can be used to visualize a vectorfield.

        NOTE: For example ArcGIS is able to visualize such 2-band raster files as a vectorfield.

        :param path_out:    the output path. If not given, it is automatically defined.
        :param fmt:         output raster format string
        :param mode:        The mode how the output is written ('uv' or 'md'; default: 'md')
                            - 'uv': outputs X-/Y shifts
                            - 'md': outputs magnitude and direction
        """
        assert mode in ['uv', 'md'], "'mode' must be either 'uv' (outputs X-/Y shifts) or 'md' " \
                                     "(outputs magnitude and direction)'. Got %s." % mode
        attr_b1 = 'X_SHIFT_M' if mode == 'uv' else 'ABS_SHIFT'
        attr_b2 = 'Y_SHIFT_M' if mode == 'uv' else 'ANGLE'

        if self.CoRegPoints_table.empty:
            raise RuntimeError('Cannot save the vector field because no tie points were found at all.')

        xshift_arr, gt, prj = points_to_raster(points=self.CoRegPoints_table['geometry'],
                                               values=self.CoRegPoints_table[attr_b1],
                                               tgt_res=self.shift.xgsd * self.grid_res,
                                               prj=self.CoRegPoints_table.crs.to_wkt(),
                                               fillVal=self.outFillVal)

        yshift_arr, gt, prj = points_to_raster(points=self.CoRegPoints_table['geometry'],
                                               values=self.CoRegPoints_table[attr_b2],
                                               tgt_res=self.shift.xgsd * self.grid_res,
                                               prj=self.CoRegPoints_table.crs.to_wkt(),
                                               fillVal=self.outFillVal)

        out_GA = GeoArray(np.dstack([xshift_arr, yshift_arr]), gt, prj, nodata=self.outFillVal)

        path_out = path_out if path_out else \
            get_generic_outpath(dir_out=os.path.join(self.dir_out, 'CoRegPoints'),
                                fName_out="CoRegVectorfield%s_ws(%s_%s)__T_%s__R_%s.tif"
                                          % (self.grid_res, self.COREG_obj.win_size_XY[0],
                                             self.COREG_obj.win_size_XY[1], self.shift.basename, self.ref.basename))

        out_GA.save(path_out, fmt=fmt if fmt else 'Gtiff')

        return out_GA

    def to_interpolated_raster(self,
                               metric: str = 'ABS_SHIFT',
                               method: str = 'RBF',
                               plot_result: bool = False,
                               lowres_spacing: int = 5,
                               v: bool = False
                               ) -> np.ndarray:
        """Interpolate the point data of the given metric into space.

        :param metric:          metric name to interpolate, i.e., one of the column names of
                                Tie_Point_Grid.CoRegPoints_table, e.g., 'ABS_SHIFT'.
        :param method:          interpolation algorithm
                                - 'RBF' (Radial Basis Function)
                                - 'GPR' (Gaussian Process Regression; equivalent to Simple Kriging)
                                - 'Kriging' (Ordinary Kriging based on pykrige)
        :param plot_result:     plot the result to assess the interpolation quality
        :param lowres_spacing:  by default, RBF, GPR, and Kriging run a lower resolution which is then linearly
                                interpolated to the full output image resolution. lowres_spacing defines the number of
                                pixels between the low resolution grid points
                                (higher values are faster but less accurate, default: 5)
        :param v:               enable verbose mode
        :return:    interpolation result as numpy array in the X/Y dimension of the target image of the co-registration
        """
        TPGI = Tie_Point_Grid_Interpolator(self, v=v)

        return TPGI.interpolate(metric=metric, method=method, plot_result=plot_result, lowres_spacing=lowres_spacing)

    @staticmethod
    def to_Raster_using_Kriging(*args, **kwargs):
        raise NotImplementedError('This method was removed in arosics version 1.8.2. To interpolate tie points into '
                                  'space, you may now use Tie_Point_Grid.to_interpolated_raster() instead.')


class Tie_Point_Refiner(object):
    """A class for performing outlier detection."""

    def __init__(self, GDF,
                 min_reliability=60,
                 rs_max_outlier: float = 10,
                 rs_tolerance: float = 2.5,
                 rs_max_iter: int = 15,
                 rs_exclude_previous_outliers: bool = True,
                 rs_timeout: float = 20,
                 rs_random_state: Optional[int] = 0,
                 q: bool = False):
        """Get an instance of Tie_Point_Refiner.

        :param GDF:                             GeoDataFrame like TiePointGrid.CoRegPoints_table containing all tie
                                                points to be filtered and the corresponding metadata
        :param min_reliability:                 minimum threshold for previously computed tie X/Y shift
                                                reliability (default: 60%)
        :param rs_max_outlier:                  RANSAC: maximum percentage of outliers to be detected
                                                (default: 10%)
        :param rs_tolerance:                    RANSAC: percentage tolerance for max_outlier_percentage
                                                (default: 2.5%)
        :param rs_max_iter:                     RANSAC: maximum iterations for finding the best RANSAC threshold
                                                (default: 15)
        :param rs_exclude_previous_outliers:    RANSAC: whether to exclude points that have been flagged as
                                                outlier by earlier filtering (default:True)
        :param rs_timeout:                      RANSAC: timeout for iteration loop in seconds (default: 20)
        :param rs_random_state:                 RANSAC random state (an integer corresponds to a fixed/pseudo-random
                                                state, None randomizes the result)

        :param q:
        """
        self.GDF = GDF.copy()
        self.min_reliability = min_reliability
        self.rs_max_outlier_percentage = rs_max_outlier
        self.rs_tolerance = rs_tolerance
        self.rs_max_iter = rs_max_iter
        self.rs_exclude_previous_outliers = rs_exclude_previous_outliers
        self.rs_timeout = rs_timeout
        self.rs_random_state = rs_random_state
        self.q = q
        self.new_cols = []
        self.ransac_model_robust = None

    def run_filtering(self, level=3):
        """Filter tie points used for shift correction.

        :param level:   tie point filter level (default: 3).
                        NOTE: lower levels are also included if a higher level is chosen

                        - Level 0: no tie point filtering
                        - Level 1: Reliablity filtering
                                   - filter all tie points out that have a low reliability according to internal tests
                        - Level 2: SSIM filtering
                                   - filters all tie points out where shift correction does not increase image
                                     similarity within matching window (measured by mean structural similarity index)
                        - Level 3: RANSAC outlier detection

        :return:
        """
        # TODO catch empty GDF

        # RELIABILITY filtering
        if level > 0:
            marked_recs = self._reliability_thresholding()  # type: Series
            self.GDF['L1_OUTLIER'] = marked_recs
            self.new_cols.append('L1_OUTLIER')

            n_flagged = len(marked_recs[marked_recs])
            perc40 = np.percentile(self.GDF.RELIABILITY, 40)

            if n_flagged / len(self.GDF) > .7:
                warnings.warn(r"More than 70%% of the found tie points have a reliability lower than %s%% and are "
                              r"therefore marked as false-positives. Consider relaxing the minimum reliability "
                              r"(parameter 'min_reliability') to avoid that. For example min_reliability=%d would only "
                              r"flag 40%% of the tie points in case of your input data."
                              % (self.min_reliability, perc40))

            if not self.q:
                print('%s tie points flagged by level 1 filtering (reliability).'
                      % n_flagged)

        # SSIM filtering
        if level > 1:
            marked_recs = self._SSIM_filtering()
            self.GDF['L2_OUTLIER'] = marked_recs  # type: Series
            self.new_cols.append('L2_OUTLIER')

            if not self.q:
                print('%s tie points flagged by level 2 filtering (SSIM).' % (len(marked_recs[marked_recs])))

        # RANSAC filtering
        if level > 2:
            # exclude previous outliers
            ransacInGDF = self.GDF[~self.GDF[self.new_cols].any(axis=1)].copy() \
                    if self.rs_exclude_previous_outliers else self.GDF

            if len(ransacInGDF) > 4:
                # running RANSAC with less than four tie points makes no sense

                marked_recs = self._RANSAC_outlier_detection(ransacInGDF)  # type: Series
                # we need to join a list here because otherwise it's merged by the 'index' column
                self.GDF['L3_OUTLIER'] = marked_recs.tolist()

                if not self.q:
                    print('%s tie points flagged by level 3 filtering (RANSAC)'
                          % (len(marked_recs[marked_recs])))
            else:
                print('RANSAC skipped because too less valid tie points have been found.')
                self.GDF['L3_OUTLIER'] = False

            self.new_cols.append('L3_OUTLIER')

        self.GDF['OUTLIER'] = self.GDF[self.new_cols].any(axis=1)
        self.new_cols.append('OUTLIER')

        return self.GDF, self.new_cols

    def _reliability_thresholding(self):
        """Exclude all records where estimated reliability of the calculated shifts is below the given threshold."""
        return self.GDF.RELIABILITY < self.min_reliability

    def _SSIM_filtering(self):
        """Exclude all records where SSIM decreased."""
        # ssim_diff  = np.median(self.GDF['SSIM_AFTER']) - np.median(self.GDF['SSIM_BEFORE'])

        # self.GDF.SSIM_IMPROVED = \
        #     self.GDF.apply(lambda GDF_row: GDF_row['SSIM_AFTER']>GDF_row['SSIM_BEFORE'] + ssim_diff, axis=1)

        return ~self.GDF.SSIM_IMPROVED

    def _RANSAC_outlier_detection(self, inGDF):
        """Detect geometric outliers between point cloud of source and estimated coordinates using RANSAC algorithm."""
        # from skimage.transform import PolynomialTransform  # import here to avoid static TLS ImportError

        src_coords = np.array(inGDF[['X_MAP', 'Y_MAP']])
        xyShift = np.array(inGDF[['X_SHIFT_M', 'Y_SHIFT_M']])
        est_coords = src_coords + xyShift

        for co, n in zip([src_coords, est_coords], ['src_coords', 'est_coords']):
            assert co.ndim == 2 and co.shape[1] == 2, "'%s' must have shape [Nx2]. Got shape %s." % (n, co.shape)

        if not 0 < self.rs_max_outlier_percentage < 100:
            raise ValueError
        min_inlier_percentage = 100 - self.rs_max_outlier_percentage

        # class PolyTF_1(PolynomialTransform):  # pragma: no cover
        #     def estimate(*data):
        #         return PolynomialTransform.estimate(*data, order=1)

        # robustly estimate affine transform model with RANSAC
        # eliminates not more than the given maximum outlier percentage of the tie points

        model_robust, inliers = None, None
        count_inliers = None
        th = 5  # start RANSAC threshold
        th_checked = {}  # dict of thresholds that already have been tried + calculated inlier percentage
        th_substract = 2
        count_iter = 0
        time_start = time()
        ideal_count = min_inlier_percentage * src_coords.shape[0] / 100

        # optimize RANSAC threshold so that it marks not much more or less than the given outlier percentage
        while True:
            if th_checked:
                th_too_strict = count_inliers < ideal_count  # True if too less inliers remaining

                # calculate new theshold using old increment
                # (but ensure th_new>0 by adjusting increment if needed)
                th_new = 0
                while th_new <= 0:
                    th_new = th + th_substract if th_too_strict else th - th_substract
                    if th_new <= 0:
                        th_substract /= 2

                # check if calculated new threshold has been used before
                th_already_checked = th_new in th_checked.keys()

                # if yes, decrease increment and recalculate new threshold
                th_substract = \
                    th_substract if not th_already_checked else \
                    th_substract / 2
                th = th_new if not th_already_checked else \
                    (th + th_substract if th_too_strict else
                     th - th_substract)

            ###############
            # RANSAC call #
            ###############

            # model_robust, inliers = ransac((src, dst),
            #                                PolynomialTransform,
            #                                min_samples=3)
            if src_coords.size and \
               est_coords.size and \
               src_coords.shape[0] > 6:
                # import here to avoid static TLS ImportError
                from skimage.measure import ransac
                from skimage.transform import AffineTransform

                model_robust, inliers = \
                    ransac((src_coords, est_coords),
                           AffineTransform,
                           min_samples=6,
                           residual_threshold=th,
                           max_trials=2000,
                           stop_sample_num=int(
                               (min_inlier_percentage - self.rs_tolerance) /
                               100 * src_coords.shape[0]
                           ),
                           stop_residuals_sum=int(
                               (self.rs_max_outlier_percentage - self.rs_tolerance) /
                               100 * src_coords.shape[0]),
                           random_state=self.rs_random_state
                           )
            else:
                warnings.warn('RANSAC filtering could not be applied '
                              'because there were too few tie points to fit a model.')
                inliers = np.array([])
                break

            count_inliers = np.count_nonzero(inliers)

            th_checked[th] = count_inliers / src_coords.shape[0] * 100
            # print(th,'\t', th_checked[th], )

            if min_inlier_percentage - self.rs_tolerance <\
               th_checked[th] <\
               min_inlier_percentage + self.rs_tolerance:
                # print('in tolerance')
                break

            if count_iter > self.rs_max_iter or \
               time() - time_start > self.rs_timeout:
                break  # keep last values and break while loop

            count_iter += 1

        outliers = inliers.__eq__(False) if inliers is not None and inliers.size else np.array([])

        if inGDF.empty or outliers is None or \
           (isinstance(outliers, list) and not outliers) or \
           (isinstance(outliers, np.ndarray) and not outliers.size):
            outseries = Series([False] * len(self.GDF))

        elif len(inGDF) < len(self.GDF):
            inGDF['outliers'] = outliers
            fullGDF = GeoDataFrame(self.GDF['POINT_ID'])
            fullGDF = fullGDF.merge(inGDF[['POINT_ID', 'outliers']],
                                    on='POINT_ID',
                                    how="outer")
            # fullGDF.outliers.copy()[~fullGDF.POINT_ID.isin(GDF.POINT_ID)] = False
            fullGDF = fullGDF.fillna(False)  # NaNs are due to exclude_previous_outliers
            outseries = fullGDF['outliers']

        else:
            outseries = Series(outliers)

        assert len(outseries) == len(self.GDF), \
            'RANSAC output validation failed.'

        self.ransac_model_robust = model_robust

        return outseries


class Tie_Point_Grid_Interpolator(object):
    """Class to interpolate tie point data into space."""

    def __init__(self, tiepointgrid: Tie_Point_Grid, v: bool = False) -> None:
        """Get an instance of Tie_Point_Grid_Interpolator.

        :param tiepointgrid:    instance of Tie_Point_Grid after computing spatial shifts
        :param v:               enable verbose mode
        """
        self.tpg = tiepointgrid
        self.v = v

    def interpolate(self,
                    metric: str,
                    method: str = 'RBF',
                    plot_result: bool = False,
                    lowres_spacing: int = 5
                    ) -> np.array:
        """Interpolate the point data of the given metric into space.

        :param metric:          metric name to interpolate, i.e., one of the column names of
                                Tie_Point_Grid.CoRegPoints_table, e.g., 'ABS_SHIFT'.
        :param method:          interpolation algorithm
                                - 'RBF' (Radial Basis Function)
                                - 'GPR' (Gaussian Process Regression; equivalent to Simple Kriging)
                                - 'Kriging' (Ordinary Kriging based on pykrige)
        :param plot_result:     plot the result to assess the interpolation quality
        :param lowres_spacing:  by default, RBF, GPR, and Kriging run a lower resolution which is then linearly
                                interpolated to the full output image resolution. lowres_spacing defines the number of
                                pixels between the low resolution grid points
                                (higher values are faster but less accurate, default: 5)
        :return:    interpolation result as numpy array in the X/Y dimension of the target image of the co-registration
        """
        t0 = time()

        rows, cols, data = self._get_pointdata(metric)
        nrows_out, ncols_out = self.tpg.shift.shape[:2]

        rows_lowres = np.arange(0, nrows_out + lowres_spacing, lowres_spacing)
        cols_lowres = np.arange(0, ncols_out + lowres_spacing, lowres_spacing)
        args = rows, cols, data, rows_lowres, cols_lowres

        if method == 'RBF':
            data_lowres = self._interpolate_via_rbf(*args)
        elif method == 'GPR':
            data_lowres = self._interpolate_via_gpr(*args)
        elif method == 'Kriging':
            data_lowres = self._interpolate_via_kriging(*args)
        else:
            raise ValueError(method)

        if lowres_spacing > 1:
            rows_full = np.arange(nrows_out)
            cols_full = np.arange(ncols_out)
            data_full = self._interpolate_regulargrid(rows_lowres, cols_lowres, data_lowres, rows_full, cols_full)
        else:
            data_full = data_lowres

        if self.v:
            print('interpolation runtime: %.2fs' % (time() - t0))
        if plot_result:
            self._plot_interpolation_result(data_full, rows, cols, data, metric)

        return data_full

    def _get_pointdata(self, metric: str):
        """Get the point data for the given metric from Tie_Point_Grid.CoRegPoints_table while ignoring outliers."""
        tiepoints = self.tpg.CoRegPoints_table[self.tpg.CoRegPoints_table.OUTLIER.__eq__(False)].copy()

        rows = np.array(tiepoints.Y_IM)
        cols = np.array(tiepoints.X_IM)
        data = np.array(tiepoints[metric])

        return rows, cols, data

    @staticmethod
    def _plot_interpolation_result(data_full: np.ndarray,
                                   rows: np.ndarray,
                                   cols: np.ndarray,
                                   data: np.ndarray,
                                   metric: str
                                   ):
        """Plot the interpolation result together with the input point data."""
        plt.figure(figsize=(7, 7))
        im = plt.imshow(data_full)
        plt.colorbar(im)
        plt.scatter(cols, rows, c=data, edgecolors='black')
        plt.title(metric)
        plt.show()

    @staticmethod
    def _interpolate_regulargrid(rows: np.ndarray,
                                 cols: np.ndarray,
                                 data: np.ndarray,
                                 rows_full: np.ndarray,
                                 cols_full: np.ndarray
                                 ):
        """Run linear regular grid interpolation."""
        RGI = RegularGridInterpolator(points=[cols, rows],
                                      values=data.T,  # must be in shape [x, y]
                                      method='linear',
                                      bounds_error=False)
        data_full = RGI(np.dstack(np.meshgrid(cols_full, rows_full)))
        return data_full

    @staticmethod
    def _interpolate_via_rbf(rows: np.ndarray,
                             cols: np.ndarray,
                             data: np.ndarray,
                             rows_full: np.ndarray,
                             cols_full: np.ndarray
                             ):
        """Run Radial Basis Function (RBF) interpolation.

        -> https://github.com/agile-geoscience/xlines/blob/master/notebooks/11_Gridding_map_data.ipynb
        -> documents the legacy scipy.interpolate.Rbf
        """
        rbf = RBFInterpolator(
            np.column_stack([cols, rows]), data,
            kernel="linear",
            # kernel="thin_plate_spline",
        )
        cols_grid, rows_grid = np.meshgrid(cols_full, rows_full)
        data_full = \
            rbf(np.column_stack([cols_grid.flat, rows_grid.flat]))\
            .reshape(rows_grid.shape)

        return data_full

    @staticmethod
    def _interpolate_via_gpr(rows: np.ndarray,
                             cols: np.ndarray,
                             data: np.ndarray,
                             rows_full: np.ndarray,
                             cols_full: np.ndarray
                             ):
        """Run Gaussian Process Regression (GPR) interpolation.

        -> https://stackoverflow.com/questions/24978052/interpolation-over-regular-grid-in-python
        """
        try:
            import sklearn  # noqa F401
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "GPR interpolation requires the optional package 'scikit-learn' to be installed. You may install it "
                "with Conda (conda install -c conda-forge scikit-learn) or Pip (pip install scikit-learn)."
            )

        from sklearn.gaussian_process.kernels import RBF
        from sklearn.gaussian_process import GaussianProcessRegressor

        gp = GaussianProcessRegressor(
            normalize_y=False,
            alpha=0.001,  # Larger values imply more input noise and result in smoother grids; default: 1e-10
            kernel=RBF(length_scale=100))
        gp.fit(np.column_stack([cols, rows]), data.T)

        cols_grid, rows_grid = np.meshgrid(cols_full, rows_full)
        data_full = \
            gp.predict(np.column_stack([cols_grid.flat, rows_grid.flat]))\
            .reshape(rows_grid.shape)

        return data_full

    @staticmethod
    def _interpolate_via_kriging(rows: np.ndarray,
                                 cols: np.ndarray,
                                 data: np.ndarray,
                                 rows_full: np.ndarray,
                                 cols_full: np.ndarray
                                 ):
        """Run Ordinary Kriging interpolation based on pykrige.

        Reference: P.K. Kitanidis, Introduction to Geostatistics: Applications in Hydrogeology,
                   (Cambridge University Press, 1997) 272 p.
        """
        try:
            import pykrige  # noqa F401
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "Ordinary Kriging requires the optional package 'pykrige' to be installed. You may install it with "
                "Conda (conda install -c conda-forge pykrige) or Pip (pip install pykrige)."
            )

        from pykrige.ok import OrdinaryKriging

        OK = OrdinaryKriging(cols.astype(float), rows.astype(float), data.astype(float),
                             variogram_model='spherical',
                             verbose=False)

        data_full, sigmasq = \
            OK.execute('grid',
                       cols_full.astype(float),
                       rows_full.astype(float),
                       backend='C',
                       # n_closest_points=12
                       )

        return data_full



# -*- coding: utf-8 -*-

# AROSICS - Automated and Robust Open-Source Image Co-Registration Software
#
# Copyright (C) 2017-2023
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import sys
import os
from typing import Union

# custom
import numpy as np
from geopandas import GeoDataFrame

# internal modules
from py_tools_ds.geo.coord_calc import calc_FullDataset_corner_positions
from py_tools_ds.geo.coord_trafo import pixelToMapYX, imYX2mapYX
from py_tools_ds.geo.raster.reproject import warp_ndarray
from geoarray import GeoArray

__author__ = 'Daniel Scheffler'


def angle_to_north(XY):
    """Calculate the angle in degrees of a given line to north in clockwise direction.

    Angle definition: between [origin:[0,0],north:[0,1]] and [origin:[0,0],pointXY:[X,Y]].
    """
    XY = np.array(XY)
    XYarr = XY if len(XY.shape) == 2 else XY.reshape((1, 2))
    return np.abs(np.degrees(np.array(np.arctan2(XYarr[:, 1], XYarr[:, 0]) - np.pi / 2)) % 360)


def get_true_corner_mapXY(fPath_or_geoarray, band=0, noDataVal=None, mp=1, v=0, q=0):  # pragma: no cover
    """Return the actual map corner coordinates of a given image file or GeoArray instance.

    :param fPath_or_geoarray:
    :param band:                <int> index of the band to be used (starting with 0)
    :param noDataVal:
    :param mp:
    :param v:
    :param q:
    :return:
    """
    # FIXME this function is not used anymore
    warnings.warn('This function is not in use anymore. Use it on your own risk!', DeprecationWarning)
    geoArr = GeoArray(fPath_or_geoarray) if not isinstance(fPath_or_geoarray, GeoArray) else fPath_or_geoarray

    rows, cols = geoArr.shape[:2]
    gt, prj = geoArr.geotransform, geoArr.projection

    assert gt and prj, 'GeoTransform an projection must be given for calculation of LonLat corner coordinates.'

    mask_1bit = np.zeros((rows, cols), dtype='uint8')  # zeros -> image area later overwritten by ones

    if noDataVal is None:
        mask_1bit[:, :] = 1
    elif noDataVal == 'ambiguous':
        warnings.warn("No data value could not be automatically detected. Thus the matching window used for shift "
                      "calculation had to be centered in the middle of the overlap area without respecting no data "
                      "values. To avoid this provide the correct no data values for reference and shift image via "
                      "'-nodata'")
        mask_1bit[:, :] = 1
    else:
        band_data = geoArr[band]  # TODO implement gdal_ReadAsArray_mp (reading in multiprocessing)
        mask_1bit[band_data != noDataVal] = 1

    if v:
        print('detected no data value', noDataVal)

    try:
        corner_coords_YX = calc_FullDataset_corner_positions(mask_1bit, assert_four_corners=False, algorithm='shapely')
    except Exception:
        if v:
            warnings.warn("\nCalculation of corner coordinates failed within algorithm 'shapely' (Exception: %s)."
                          " Using algorithm 'numpy' instead." % sys.exc_info()[1])
        # FIXME numpy algorithm returns wrong values for S2A_OPER_MSI_L1C_TL_SGS__20160608T153121_A005024_T33UUU_B03.jp2
        # FIXME (Hannes)
        corner_coords_YX = \
            calc_FullDataset_corner_positions(mask_1bit, assert_four_corners=False, algorithm='numpy')

    if len(corner_coords_YX) == 4:  # this avoids shapely self intersection
        corner_coords_YX = list(np.array(corner_coords_YX)[[0, 1, 3, 2]])  # UL, UR, LL, LR => UL, UR, LR, LL

    # check if enough unique coordinates have been found
    if not len(GeoDataFrame(corner_coords_YX).drop_duplicates().values) >= 3:
        if not q:
            warnings.warn('\nThe algorithm for automatically detecting the actual image coordinates did not find '
                          'enough unique corners. Using outer image corner coordinates instead.')
        corner_coords_YX = ((0, 0), (0, cols - 1), (rows - 1, 0), (rows - 1, cols - 1))

    # check if all points are unique
    # all_coords_are_unique = len([UL, UR, LL, LR]) == len(GeoDataFrame([UL, UR, LL, LR]).drop_duplicates().values)
    # UL, UR, LL, LR = \
    #   (UL, UR, LL, LR) if all_coords_are_unique else ((0, 0), (0, cols-1), (rows-1, 0), (rows-1, cols-1))

    def get_mapYX(YX): return pixelToMapYX(list(reversed(YX)), geotransform=gt, projection=prj)[0]
    corner_pos_XY = [list(reversed(i)) for i in [get_mapYX(YX) for YX in corner_coords_YX]]
    return corner_pos_XY


def get_subset_GeoTransform(gt_fullArr, subset_box_imYX):
    gt_subset = list(gt_fullArr[:])  # copy
    gt_subset[3], gt_subset[0] = imYX2mapYX(subset_box_imYX[0], gt_fullArr)
    return gt_subset


def get_gdalReadInputs_from_boxImYX(boxImYX):
    """Return row_start,col_start,rows_count,cols_count and assumes boxImYX as [UL_YX,UR_YX,LR_YX,LL_YX)."""
    rS, cS = boxImYX[0]
    clip_sz_x = abs(boxImYX[1][1] - boxImYX[0][1])  # URx-ULx
    clip_sz_y = abs(boxImYX[0][0] - boxImYX[3][0])  # ULy-LLy
    return cS, rS, clip_sz_x, clip_sz_y


def get_GeoArrayPosition_from_boxImYX(boxImYX):
    """Return row_start,row_end,col_start,col_end and assumes boxImYX as [UL_YX,UR_YX,LR_YX,LL_YX)."""
    rS, cS = boxImYX[0]  # UL
    rE, cE = boxImYX[2]  # LR
    return rS, rE - 1, cS, cE - 1  # -1 because boxImYX represents outer box and includes the LR corner of LR pixel


def has_metaRotation(path_or_geoarray: Union[GeoArray, str]):
    """Return True if there is a row or column rotation due to the given GDAL GeoTransform tuple."""
    gt = GeoArray(path_or_geoarray).gt

    return gt[2] or gt[4]


def remove_metaRotation(gA_rot: GeoArray, rspAlg='cubic') -> GeoArray:
    """Remove any metadata rotation (a rotation that only exists in the map info)."""
    gA = GeoArray(*warp_ndarray(gA_rot[:], gA_rot.gt, gA_rot.prj,
                                rspAlg=rspAlg,
                                # out_gsd=(gA_rot.xgsd, gA_rot.ygsd)
                                ),
                  nodata=gA_rot.nodata)
    gA.basename = os.path.basename(gA.basename)
    gA.meta = gA.meta

    return gA



# -*- coding: utf-8 -*-

# AROSICS - Automated and Robust Open-Source Image Co-Registration Software
#
# Copyright (C) 2017-2023
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import os
from copy import copy
from typing import Tuple, Union, Optional
from collections import OrderedDict

# custom
from osgeo import gdal
from packaging.version import parse as parse_version
try:
    import pyfftw
    # pyfftw>=0.13.0 is currently not used due to https://github.com/pyFFTW/pyFFTW/issues/294
    if parse_version(pyfftw.__version__) >= parse_version('0.13.0'):
        pyfftw = None
except ImportError:
    pyfftw = None
import numpy as np

from matplotlib import pyplot as plt  # noqa F401
from geopandas import GeoDataFrame  # noqa F401

from .Tie_Point_Grid import Tie_Point_Grid
from .CoReg import COREG
from .DeShifter import DESHIFTER
from .geometry import has_metaRotation, remove_metaRotation
from py_tools_ds.geo.coord_trafo import transform_any_prj, reproject_shapelyGeometry
from py_tools_ds.geo.map_info import geotransform2mapinfo
from geoarray import GeoArray

__author__ = 'Daniel Scheffler'


class COREG_LOCAL(object):
    """
    COREG_LOCAL applies the algorithm to detect spatial shifts to the whole overlap area of the input images.

    Spatial shifts are calculated for each point in grid of which the parameters can be adjusted using keyword
    arguments. Shift correction performs a polynomial transformation using the calculated shifts of each point in the
    grid as GCPs. Thus this class can be used to correct for locally varying geometric distortions of the target image.

    See help(COREG_LOCAL) for documentation.
    """

    def __init__(self,
                 im_ref: Union[GeoArray, str],
                 im_tgt: Union[GeoArray, str],
                 grid_res: float,
                 max_points: int = None,
                 window_size: Tuple[int, int] = (256, 256),
                 path_out: str = None,
                 fmt_out: str = 'ENVI',
                 out_crea_options: list = None,
                 projectDir: str = None,
                 r_b4match: int = 1,
                 s_b4match: int = 1,
                 max_iter: int = 5,
                 max_shift: int = 5,
                 tieP_filter_level: int = 3,
                 min_reliability: float = 60,
                 rs_max_outlier: float = 10,
                 rs_tolerance: float = 2.5,
                 rs_random_state: Optional[int] = 0,
                 align_grids: bool = True,
                 match_gsd: bool = False,
                 out_gsd: float = None,
                 target_xyGrid=None,
                 resamp_alg_deshift: str = 'cubic',
                 resamp_alg_calc: str = 'cubic',
                 footprint_poly_ref: str = None,
                 footprint_poly_tgt: str = None,
                 data_corners_ref: list = None,
                 data_corners_tgt: list = None,
                 outFillVal: int = -9999,
                 nodata: Tuple[int, int] = (None, None),
                 calc_corners: bool = True,
                 binary_ws: bool = True,
                 force_quadratic_win: bool = True,
                 mask_baddata_ref: Union[GeoArray, str] = None,
                 mask_baddata_tgt: Union[GeoArray, str] = None,
                 CPUs: int = None,
                 progress: bool = True,
                 v: bool = False,
                 q: bool = False,
                 ignore_errors: bool = True
                 ) -> None:
        """
        Get an instance of COREG_LOCAL.

        :param im_ref:
            source path of reference image (any GDAL compatible image format is supported)

        :param im_tgt:
            source path of image to be shifted (any GDAL compatible image format is supported)

        :param grid_res:
            tie point grid resolution in pixels of the target image (x-direction)

        :param max_points:
            maximum number of points used to find coregistration tie points

            NOTE: Points are selected randomly from the given point grid (specified by 'grid_res').
                  If the point grid does not provide enough points, all available points are chosen.

        :param window_size:
            custom matching window size [pixels] (default: (256,256))

        :param path_out:
            target path of the coregistered image
            - if None (default), no output is written to disk
            - if 'auto': /dir/of/im1/<im1>__shifted_to__<im0>.bsq

        :param fmt_out:
            raster file format for output file. ignored if path_out is None. Can be any GDAL compatible raster file
            format (e.g. 'ENVI', 'GTIFF'; default: ENVI). Refer to https://gdal.org/drivers/raster/index.html to get a
            full list of supported formats.

        :param out_crea_options:
            GDAL creation options for the output image, e.g. ["QUALITY=80", "REVERSIBLE=YES", "WRITE_METADATA=YES"]

        :param projectDir:
            name of a project directory where to store all the output results. If given, name is inserted into all
            automatically generated output paths.

        :param r_b4match:
            band of reference image to be used for matching (starts with 1; default: 1)

        :param s_b4match:
            band of shift image to be used for matching (starts with 1; default: 1)

        :param max_iter:
            maximum number of iterations for matching (default: 5)

        :param max_shift:
            maximum shift distance in reference image pixel units (default: 5 px)

        :param tieP_filter_level:
            filter tie points used for shift correction in different levels (default: 3).
            NOTE: lower levels are also included if a higher level is chosen

            - Level 0: no tie point filtering
            - Level 1: Reliablity filtering
                       - filter all tie points out that have a low reliability according to internal tests
            - Level 2: SSIM filtering
                       - filters all tie points out where shift correction does not increase image similarity within
                         matching window (measured by mean structural similarity index)
            - Level 3: RANSAC outlier detection

        :param min_reliability:
            Tie point filtering: minimum reliability threshold, below which tie points are marked as false-positives
            (default: 60%)
            - accepts values between 0% (no reliability) and 100 % (perfect reliability)
            HINT: decrease this value in case of poor signal-to-noise ratio of your input data

        :param rs_max_outlier:
            RANSAC tie point filtering: proportion of expected outliers (default: 10%)

        :param rs_tolerance:
            RANSAC tie point filtering: percentage tolerance for max_outlier_percentage (default: 2.5%)

        :param rs_random_state:
            RANSAC random state (an integer corresponds to a fixed/pseudo-random state, None randomizes the result)

        :param align_grids:
            True: align the input coordinate grid to the reference (does not affect the output pixel size as long as
            input and output pixel sizes are compatible (5:30 or 10:30 but not 4:30), default = True

        :param match_gsd:
            True: match the input pixel size to the reference pixel size,
            default = False

        :param out_gsd:
            output pixel size in units of the reference coordinate system (default = pixel size of the input array),
            given values are overridden by match_gsd=True

        :param target_xyGrid:
            a list with a target x-grid and a target y-grid like [[15,45], [15,45]]
            This overrides 'out_gsd', 'align_grids' and 'match_gsd'.

        :param resamp_alg_deshift:
            the resampling algorithm to be used for shift correction (if neccessary)
            valid algorithms: nearest, bilinear, cubic, cubic_spline, lanczos, average, mode, max, min, med, q1, q3
            (default: cubic)

        :param resamp_alg_calc:
            the resampling algorithm to be used for all warping processes during calculation of spatial shifts
            valid algorithms: nearest, bilinear, cubic, cubic_spline, lanczos, average, mode, max, min, med, q1, q3
            (default: cubic (highly recommended))

        :param footprint_poly_ref:
            footprint polygon of the reference image (WKT string or shapely.geometry.Polygon),
            e.g. 'POLYGON ((299999 6000000, 299999 5890200, 409799 5890200, 409799 6000000, 299999 6000000))'

        :param footprint_poly_tgt:
            footprint polygon of the image to be shifted (WKT string or shapely.geometry.Polygon)
            e.g. 'POLYGON ((299999 6000000, 299999 5890200, 409799 5890200, 409799 6000000, 299999 6000000))'

        :param data_corners_ref:
            map coordinates of data corners within reference image. ignored if footprint_poly_ref is given.

        :param data_corners_tgt:
            map coordinates of data corners within image to be shifted. ignored if footprint_poly_tgt is given.

        :param outFillVal:
            if given the generated tie point grid is filled with this value in case no match could be found during
            co-registration (default: -9999)

        :param nodata:
            no data values for reference image and image to be shifted

        :param calc_corners:
            calculate true positions of the dataset corners in order to get a useful matching window position within
            the actual image overlap
            (default: True; deactivated if 'data_corners_im0' and 'data_corners_im1' are given)

        :param binary_ws:
            use binary X/Y dimensions for the matching window (default: True)

        :param force_quadratic_win:
            force a quadratic matching window (default: True)

        :param mask_baddata_ref:
            path to a 2D boolean mask file (or an instance of BadDataMask) for the reference image where all bad data
            pixels (e.g. clouds) are marked with True and the remaining pixels with False. Must have the same
            geographic extent and projection like 'im_ref'. The mask is used to check if the chosen matching window
            position is valid in the sense of useful data. Otherwise this window position is rejected.

        :param mask_baddata_tgt:
            path to a 2D boolean mask file (or an instance of BadDataMask) for the image to be shifted where all bad
            data pixels (e.g. clouds) are marked with True and the remaining pixels with False. Must have the same
            geographic extent and projection like 'im_ref'. The mask is used to check if the chosen matching window
            position is valid in the sense of useful data. Otherwise this window position is rejected.

        :param CPUs:
            number of CPUs to use during calculation of tie point grid (default: None, which means 'all CPUs available')

        :param progress:
            show progress bars (default: True)

        :param v:
            verbose mode (default: False)

        :param q:
            quiet mode (default: False)

        :param ignore_errors:
            Useful for batch processing. (default: False)
        """
        # assertions / input validation
        assert gdal.GetDriverByName(fmt_out), "'%s' is not a supported GDAL driver." % fmt_out
        if match_gsd and out_gsd:
            warnings.warn("'-out_gsd' is ignored because '-match_gsd' is set.\n")
        if out_gsd:
            assert isinstance(out_gsd, list) and len(out_gsd) == 2, 'out_gsd must be a list with two values.'

        self.params = dict([x for x in locals().items() if x[0] != "self" and not x[0].startswith('__')])

        # NOTE: self.imref and self.im2shift are handled completely independent from self.COREG_obj.ref and
        #       self.COREG_obj.shift. self.COREG_obj.ref and self.COREG_obj.shift are used for shift calculation and
        #       correction is applied to self.im2shift.
        self.imref = GeoArray(im_ref, nodata=nodata[0], progress=progress, q=q)
        self.im2shift = GeoArray(im_tgt, nodata=nodata[1], progress=progress, q=q)
        self.path_out = path_out  # updated by self.set_outpathes
        self.fmt_out = fmt_out
        self.out_creaOpt = out_crea_options
        self._projectDir = projectDir
        self.grid_res = grid_res
        self.max_points = max_points
        self.window_size = window_size
        self.max_shift = max_shift
        self.max_iter = max_iter
        self.tieP_filter_level = tieP_filter_level
        self.min_reliability = min_reliability
        self.rs_max_outlier = rs_max_outlier
        self.rs_tolerance = rs_tolerance
        self.rs_random_state = rs_random_state
        self.align_grids = align_grids
        self.match_gsd = match_gsd
        self.out_gsd = out_gsd
        self.target_xyGrid = target_xyGrid
        self.rspAlg_DS = resamp_alg_deshift  # TODO convert integers to strings
        self.rspAlg_calc = resamp_alg_calc
        self.calc_corners = calc_corners
        self.nodata = nodata
        self.outFillVal = outFillVal
        self.bin_ws = binary_ws
        self.force_quadratic_win = force_quadratic_win
        self.CPUs = CPUs
        self.path_verbose_out = ''  # TODO
        self.v = v
        self.q = q if not v else False  # overridden by v
        self.progress = progress if not q else False  # overridden by v
        self.ignErr = ignore_errors  # FIXME this is not yet implemented for COREG_LOCAL

        assert self.tieP_filter_level in range(4), 'Invalid tie point filter level.'
        assert isinstance(self.imref, GeoArray) and isinstance(self.im2shift, GeoArray), \
            'Something went wrong with the creation of GeoArray instances for reference or target image. The created ' \
            'instances do not seem to belong to the GeoArray class. If you are working in Jupyter Notebook, reset ' \
            'the kernel and try again.'

        COREG.__dict__['_set_outpathes'](self, self.imref, self.im2shift)
        # make sure that the output directory of coregistered image is the project directory if a project directory is
        # given
        if path_out and projectDir and os.path.basename(self.path_out):
            self.path_out = os.path.join(self.projectDir, os.path.basename(self.path_out))

        gdal.AllRegister()

        # resample input data in case there is a metadata rotation (not handled by AROSICS)
        self._check_and_handle_metaRotation()

        try:
            # ignore_errors must be False because in case COREG init fails, coregistration for the whole scene fails
            self.COREG_obj = COREG(self.imref, self.im2shift,
                                   ws=window_size,
                                   footprint_poly_ref=footprint_poly_ref,
                                   footprint_poly_tgt=footprint_poly_tgt,
                                   data_corners_ref=data_corners_ref,
                                   data_corners_tgt=data_corners_tgt,
                                   resamp_alg_calc=self.rspAlg_calc,
                                   calc_corners=calc_corners,
                                   r_b4match=r_b4match,
                                   s_b4match=s_b4match,
                                   max_iter=max_iter,
                                   max_shift=max_shift,
                                   nodata=nodata,
                                   mask_baddata_ref=None,  # see below
                                   mask_baddata_tgt=None,
                                   CPUs=self.CPUs,
                                   force_quadratic_win=self.force_quadratic_win,
                                   binary_ws=self.bin_ws,
                                   progress=self.progress,
                                   v=v,
                                   q=q,
                                   ignore_errors=False)
        except Exception:
            warnings.warn('\nFirst attempt to check the functionality of co-registration failed. Check your '
                          'input data and parameters. The following error occurred:', stacklevel=3)
            raise

        if pyfftw:
            self.check_if_fftw_works()

        # add bad data mask
        # (mask is not added during initialization of COREG object in order to avoid bad data area errors there)
        if mask_baddata_ref is not None:
            self.COREG_obj.ref.mask_baddata = mask_baddata_ref
        if mask_baddata_tgt is not None:
            self.COREG_obj.shift.mask_baddata = mask_baddata_tgt

        self._tiepoint_grid = None  # set by self.tiepoint_grid
        self._CoRegPoints_table = None  # set by self.CoRegPoints_table
        self._coreg_info = None  # set by self.coreg_info
        self.deshift_results = None  # set by self.correct_shifts()
        self._success = None  # set by self.success property

    def _check_and_handle_metaRotation(self):
        """Check if the provided input data have a metadata rotation and if yes, correct it AND equalize grids.

        In case there is a rotation, the GDAL GeoTransform is not 0 at positions 2 or 4. So far, AROSICS does not
        handle such rotations, so the resampling is needed to make things work. The pixel grid equalization is also
        done here to avoid a double-resampling (grid would be equalized by COREG.equalize_pixGrids() otherwise).
        """
        grid2use = 'ref' if self.im2shift.xgsd <= self.imref.xgsd else 'shift'

        if has_metaRotation(self.imref) or has_metaRotation(self.im2shift):

            msg = 'The %s image needs to be resampled because it has a row/column rotation in '\
                  'its map info which is not handled by AROSICS.'

            if grid2use == 'ref':
                if has_metaRotation(self.imref):
                    warnings.warn(msg % 'reference')
                    self.imref = remove_metaRotation(self.imref)

                # resample target to reference image
                if not self.q:
                    print('Adapting the target image pixel grid to the one of the reference image for shift detection.')
                self.im2shift.reproject_to_new_grid(prototype=self.imref, CPUs=self.CPUs)

            else:
                # remove any metadata rotation (a rotation that only exists in the map info)
                if has_metaRotation(self.im2shift):
                    warnings.warn(msg % 'target')
                    self.im2shift = remove_metaRotation(self.im2shift)

                # resample reference to target image
                print('Adapting the reference image pixel grid to the one of the target image for shift detection.')
                self.imref.reproject_to_new_grid(prototype=self.im2shift, CPUs=self.CPUs)

    def check_if_fftw_works(self) -> None:
        """Assign the attribute 'fftw_works' to self.COREG_obj by executing shift calculation once with muted output."""
        # calculate global shift once in order to check is fftw works
        try:
            self.COREG_obj.q = True
            self.COREG_obj.v = False
            self.COREG_obj.calculate_spatial_shifts()
        except RuntimeError:
            if self.COREG_obj.fftw_works is not None:
                pass
            else:
                warnings.warn('\nFirst attempt to check if functionality of co-registration failed. Check your '
                              'input data and parameters. The following error occurred:', stacklevel=3)
                raise

        self.COREG_obj.q = self.q
        self.COREG_obj.v = self.v

    @property
    def projectDir(self) -> str:
        if self._projectDir:
            if len(os.path.split(self._projectDir)) == 1:
                return os.path.abspath(os.path.join(os.path.curdir, self._projectDir))
            else:
                return os.path.abspath(self._projectDir)
        else:
            # return a project name that not already has a corresponding folder on disk
            root_dir = os.path.dirname(self.im2shift.filePath) if self.im2shift.filePath else os.path.curdir
            fold_name = 'UntitledProject_1'

            while os.path.isdir(os.path.join(root_dir, fold_name)):
                fold_name = '%s_%s' % (fold_name.split('_')[0], int(fold_name.split('_')[-1]) + 1)

            self._projectDir = os.path.join(root_dir, fold_name)
            return self._projectDir

    @property
    def tiepoint_grid(self) -> Tie_Point_Grid:
        if self._tiepoint_grid:
            return self._tiepoint_grid
        else:
            self.calculate_spatial_shifts()
            return self._tiepoint_grid

    @property
    def CoRegPoints_table(self) -> GeoDataFrame:
        """Return a GeoDataFrame containing all the results from coregistration for all points in the tie point grid.

        Columns of the GeoDataFrame: 'geometry','POINT_ID','X_IM','Y_IM','X_MAP','Y_MAP','X_WIN_SIZE', 'Y_WIN_SIZE',
                                     'X_SHIFT_PX','Y_SHIFT_PX', 'X_SHIFT_M', 'Y_SHIFT_M', 'ABS_SHIFT' and 'ANGLE'
        """
        return self.tiepoint_grid.CoRegPoints_table

    @property
    def success(self) -> bool:
        self._success = self.tiepoint_grid.GCPList != []
        return self._success

    def calculate_spatial_shifts(self) -> None:
        self._tiepoint_grid = \
            Tie_Point_Grid(self.COREG_obj, self.grid_res,
                           max_points=self.max_points,
                           outFillVal=self.outFillVal,
                           resamp_alg_calc=self.rspAlg_calc,
                           tieP_filter_level=self.tieP_filter_level,
                           outlDetect_settings=dict(
                               min_reliability=self.min_reliability,
                               rs_max_outlier=self.rs_max_outlier,
                               rs_tolerance=self.rs_tolerance,
                               rs_random_state=self.rs_random_state),
                           dir_out=self.projectDir,
                           CPUs=self.CPUs,
                           progress=self.progress,
                           v=self.v,
                           q=self.q)
        self._tiepoint_grid.get_CoRegPoints_table()

        if self.v:
            print('Visualizing CoReg points grid...')
            self.view_CoRegPoints(figsize=(10, 10))

    def show_image_footprints(self):
        """Show a web map containing the calculated footprints and overlap area of the input images.

        NOTE: This method is intended to be called from Jupyter Notebook.
        """
        return self.COREG_obj.show_image_footprints()

    def view_CoRegPoints(self,
                         shapes2plot: str = 'points',
                         attribute2plot: str = 'ABS_SHIFT',
                         cmap: plt.cm = None,
                         exclude_fillVals: bool = True,
                         backgroundIm: str = 'tgt',
                         hide_filtered: bool = True,
                         figsize: tuple = None,
                         figsize_multiplier: float = 1,
                         title: str = '',
                         vector_scale: float = 1.,
                         savefigPath: str = '',
                         savefigDPI: int = 96,
                         showFig: bool = True,
                         vmin: float = None,
                         vmax: float = None,
                         return_map: bool = False
                         ) -> Optional[Tuple]:
        """
        Show a map of the calculated tie point grid with the target image as background.

        :param shapes2plot:         'points': plot points representing values of 'attribute2plot' onto the map
                                    'vectors': plot shift vectors onto the map
        :param attribute2plot:      the attribute of the tie point grid to be shown (default: 'ABS_SHIFT')
        :param cmap:                a custom color map to be applied to the plotted grid points (default: 'RdYlGn_r')
        :param exclude_fillVals:    whether to exclude those points of the grid where spatial shift detection failed
        :param backgroundIm:        whether to use the target or the reference image as map background. Possible
                                    options are 'ref' and 'tgt' (default: 'tgt')
        :param hide_filtered:       hide all points that have been filtered out according to tie point filter level
        :param figsize:             size of the figure to be viewed, e.g. (10, 10); automatically estimated if not given
        :param figsize_multiplier:  if given, the figure size is multiplied with this value
        :param title:               plot title
        :param vector_scale:        scale factor for shift vector length (default: 1 -> no scaling)
        :param savefigPath:         path where to save the figure
        :param savefigDPI:          DPI resolution of the output figure when saved to disk (default: 96)
        :param showFig:             whether to show or to hide the figure (default: True)
        :param vmin:                minimum value of 'attribute2plot' to be included in the figure
        :param vmax:                maximum value of 'attribute2plot' to be included in the figure
        :param return_map:          whether to return the figure and axis objects (default: False)
        :return:    tuple of figure and axis objects or None in case return_map is set to False
        """
        from matplotlib import pyplot as plt  # noqa
        from matplotlib.offsetbox import AnchoredText
        from cartopy.crs import PlateCarree
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        # get background image (reference or target image)
        if backgroundIm not in ['tgt', 'ref']:
            raise ValueError('backgroundIm')
        backgroundIm = self.im2shift if backgroundIm == 'tgt' else self.imref

        # make sure the output figure has a reasonable size, also if figsize is not given
        if not figsize:
            r, c = backgroundIm.shape[:2]
            figsize = (8, r / c * 8) if r > c else (c / r * 8, 8)

        # apply figsize multiplier
        if figsize_multiplier:
            if figsize_multiplier < 0:
                raise ValueError(figsize_multiplier, 'The figure size multiplier must be a positive finite number.')
            figsize = (figsize[0] * figsize_multiplier, figsize[1] * figsize_multiplier)

        # get a map showing the background image
        fig, ax = backgroundIm.show_map(figsize=figsize,
                                        nodataVal=self.nodata[1],
                                        return_map=True,
                                        band=self.COREG_obj.shift.band4match)

        # set figure title
        dict_attr_title = dict(
            X_WIN_SIZE='size of the matching window in x-direction [pixels]',
            Y_WIN_SIZE='size of the matching window in y-direction [pixels]',
            X_SHIFT_PX='absolute shifts in x-direction [pixels]',
            Y_SHIFT_PX='absolute shifts in y-direction [pixels]',
            X_SHIFT_M='absolute shifts in x-direction [map units]',
            Y_SHIFT_M='absolute shifts in y-direction [map units]',
            ABS_SHIFT='absolute shift vector length [map units]',
            ANGLE='shift vector direction [angle in degrees]',
            SSIM_BEFORE='structural similarity index before co-registration',
            SSIM_AFTER='structural similarity index after co-registration',
            SSIM_IMPROVED='structural similarity index improvement through co-registration [yes/no]',
            RELIABILITY='reliability of the computed shift vector'
        )
        if title:
            ax.set_title(title)
        elif attribute2plot in dict_attr_title:
            ax.set_title(dict_attr_title[attribute2plot], pad=20)
        elif attribute2plot in self.CoRegPoints_table.columns:
            ax.set_title(attribute2plot)
        else:
            raise ValueError(attribute2plot, "Invalid value for 'attribute2plot'. Valid values are: %s."
                             % ", ".join(self.CoRegPoints_table.columns))

        if not self.CoRegPoints_table.empty:
            # get GeoDataFrame containing everything needed for plotting
            outlierCols = [c for c in self.CoRegPoints_table.columns if 'OUTLIER' in c]
            attr2include = ['geometry', attribute2plot] + outlierCols + ['X_SHIFT_M', 'Y_SHIFT_M']
            GDF = self.CoRegPoints_table.loc[self.CoRegPoints_table.X_SHIFT_M != self.outFillVal, attr2include].copy()\
                if exclude_fillVals else self.CoRegPoints_table.loc[:, attr2include]

            # get LonLat coordinates for all points
            XY = np.array([geom.coords.xy for geom in GDF.geometry]).reshape(-1, 2)
            lon, lat = transform_any_prj(self.im2shift.projection, 4326, XY[:, 0], XY[:, 1])
            GDF['Lon'], GDF['Lat'] = lon, lat

            # get colors for all points
            palette = cmap if cmap is not None else plt.colormaps.get_cmap('RdYlGn_r')
            if cmap is None and attribute2plot == 'ANGLE':
                import cmocean
                palette = getattr(cmocean.cm, 'delta')

            if hide_filtered:
                if self.tieP_filter_level > 0:
                    GDF = GDF[GDF.L1_OUTLIER.__eq__(False)].copy()
                if self.tieP_filter_level > 1:
                    GDF = GDF[GDF.L2_OUTLIER.__eq__(False)].copy()
                if self.tieP_filter_level > 2:
                    GDF = GDF[GDF.L3_OUTLIER.__eq__(False)].copy()
            else:
                marker = 'o' if len(GDF) < 10000 else '.'
                common_kw = dict(marker=marker, alpha=1.0, transform=PlateCarree())
                if self.tieP_filter_level > 0:
                    # flag level 1 outliers
                    GDF_filt = GDF[GDF.L1_OUTLIER.__eq__(True)].copy()
                    ax.scatter(GDF_filt['Lon'], GDF_filt['Lat'], c='b', s=250, label='reliability',
                               **common_kw)
                if self.tieP_filter_level > 1:
                    # flag level 2 outliers
                    GDF_filt = GDF[GDF.L2_OUTLIER.__eq__(True)].copy()
                    ax.scatter(GDF_filt['Lon'], GDF_filt['Lat'], c='r', s=150, label='SSIM',
                               **common_kw)
                if self.tieP_filter_level > 2:
                    # flag level 3 outliers
                    GDF_filt = GDF[GDF.L3_OUTLIER.__eq__(True)].copy()
                    ax.scatter(GDF_filt['Lon'], GDF_filt['Lat'], c='y', s=250, label='RANSAC',
                               **common_kw)

                if self.tieP_filter_level > 0:
                    ax.legend(loc=0, scatterpoints=1)

            # plot all points or vectors on top
            if not GDF.empty:
                vmin_auto, vmax_auto = \
                    (np.percentile(GDF[attribute2plot], 0),
                     np.percentile(GDF[attribute2plot], 98)) \
                    if attribute2plot != 'ANGLE' else (0, 360)
                vmin = vmin if vmin is not None else vmin_auto
                vmax = vmax if vmax is not None else vmax_auto

                if shapes2plot == 'vectors':
                    # plot shift vectors
                    # doc: https://matplotlib.org/devdocs/api/_as_gen/matplotlib.axes.Axes.quiver.html
                    mappable = ax.quiver(
                        GDF['Lon'].values, GDF['Lat'].values,
                        -GDF['X_SHIFT_M'].values,
                        -GDF['Y_SHIFT_M'].values,  # invert absolute shifts to make arrows point to tgt
                        GDF[attribute2plot].clip(vmin, vmax),  # sets the colors
                        scale=1200 / vector_scale,  # larger values decrease the arrow length
                        width=.0015,  # arrow width (in relation to plot width)
                        # linewidth=1, # maybe use this to mark outliers instead of scatter points
                        cmap=palette,
                        pivot='middle',  # position the middle point of the arrows onto the tie point location
                        transform=PlateCarree()
                        )

                elif shapes2plot == 'points':
                    # plot tie points
                    mappable = ax.scatter(
                        GDF['Lon'], GDF['Lat'],
                        c=GDF[attribute2plot],
                        lw=0,
                        cmap=palette,
                        marker='o' if len(GDF) < 10000 else '.',
                        s=50,
                        alpha=1.0,
                        vmin=vmin,
                        vmax=vmax,
                        transform=PlateCarree())
                    pass
                else:
                    raise ValueError("The parameter 'shapes2plot' must be set to 'vectors' or 'points'. "
                                     "Received %s." % shapes2plot)

                # add colorbar
                divider = make_axes_locatable(ax)
                cax = divider.new_vertical(size="2%", pad=0.4, pack_start=True,
                                           axes_class=plt.Axes  # needed because ax is a GeoAxis instance
                                           )
                fig.add_axes(cax)
                fig.colorbar(mappable, cax=cax, orientation="horizontal")

                # hack to enlarge the figure on the top to avoid cutting off the title (everthing else has no effect)
                divider.new_vertical(size="2%", pad=0.4, pack_start=False, axes_class=plt.Axes)

            else:
                msg = "The map does not contain any tie points \n" \
                      "because all the found tie points were flagged as false-positives."
                ax.add_artist(AnchoredText(msg, loc='lower center', prop=dict(c='r')))

                if not self.q:
                    warnings.warn(msg)

        else:
            msg = "The map does not contain any tie points because no tie points were found at all."
            ax.add_artist(AnchoredText(msg, loc='lower center', prop=dict(c='r')))

            if not self.q:
                warnings.warn(msg)

        # remove white space around the figure
        plt.subplots_adjust(top=.95, bottom=.05, right=.95, left=.05)

        if savefigPath:
            fig.savefig(savefigPath, dpi=savefigDPI, pad_inches=0.1, bbox_inches='tight')

        if return_map:
            return fig, ax

        if showFig and not self.q:
            plt.show(block=True)
        else:
            plt.close(fig)

    def view_CoRegPoints_folium(self, attribute2plot: str = 'ABS_SHIFT'):
        warnings.warn(UserWarning('This function is still under construction and may not work as expected!'))
        assert self.CoRegPoints_table is not None, 'Calculate tie point grid first!'

        import folium
        import geojson
        from folium.raster_layers import ImageOverlay

        lon_min, lat_min, lon_max, lat_max = \
            reproject_shapelyGeometry(self.im2shift.box.mapPoly, self.im2shift.projection, 4326).bounds
        center_lon, center_lat = (lon_min + lon_max) / 2, (lat_min + lat_max) / 2

        # get image to plot
        image2plot = self.im2shift[:, :, 0]  # FIXME hardcoded band

        from py_tools_ds.geo.raster.reproject import warp_ndarray
        image2plot, gt, prj = \
            warp_ndarray(image2plot, self.im2shift.geotransform, self.im2shift.projection,
                         in_nodata=self.nodata[1], out_nodata=self.nodata[1], out_XYdims=(1000, 1000), q=True,
                         out_prj='epsg:3857')  # image must be transformed into web mercator projection

        # create map
        map_osm = folium.Map(location=[center_lat, center_lon])  # ,zoom_start=3)
        # import matplotlib
        ImageOverlay(
            colormap=lambda x: (1, 0, 0, x),  # TODO a colormap must be given
            # colormap=matplotlib.cm.gray, # does not work
            image=image2plot, bounds=[[lat_min, lon_min], [lat_max, lon_max]],
        ).add_to(map_osm)

        points_values = self.CoRegPoints_table[['geometry', attribute2plot]]
        points_values.geometry.crs = points_values.crs
        folium.GeoJson(points_values).add_to(map_osm)

        # add overlap polygon
        overlapPoly = reproject_shapelyGeometry(self.COREG_obj.overlap_poly, self.im2shift.prj, 4326)
        gjs = geojson.Feature(geometry=overlapPoly, properties={})
        folium.GeoJson(gjs).add_to(map_osm)

        return map_osm

    def _get_updated_map_info_meanShifts(self) -> list:
        """Return the updated map info of the target image, shifted on the basis of the mean X/Y shifts."""
        original_map_info = geotransform2mapinfo(self.im2shift.gt, self.im2shift.prj)
        updated_map_info = copy(original_map_info)
        updated_map_info[3] = str(float(original_map_info[3]) + self.tiepoint_grid.mean_x_shift_map)
        updated_map_info[4] = str(float(original_map_info[4]) + self.tiepoint_grid.mean_y_shift_map)
        return updated_map_info

    @property
    def coreg_info(self) -> dict:
        """Return a dictionary containing everthing to correct the detected local displacements of the target image."""
        if self._coreg_info:
            return self._coreg_info
        else:
            if not self._tiepoint_grid:
                self.calculate_spatial_shifts()

            TPG = self._tiepoint_grid

            self._coreg_info = {
                'GCPList': TPG.GCPList,
                'mean_shifts_px': {'x': TPG.mean_x_shift_px if TPG.GCPList else None,
                                   'y': TPG.mean_y_shift_px if TPG.GCPList else None},
                'mean_shifts_map': {'x': TPG.mean_x_shift_map if TPG.GCPList else None,
                                    'y': TPG.mean_y_shift_map if TPG.GCPList else None},
                'updated map info means': self._get_updated_map_info_meanShifts() if TPG.GCPList else None,
                'original map info': geotransform2mapinfo(self.imref.gt, self.imref.prj),
                'reference projection': self.imref.prj,
                'reference geotransform': self.imref.gt,
                'reference grid': [[self.imref.gt[0], self.imref.gt[0] + self.imref.gt[1]],
                                   [self.imref.gt[3], self.imref.gt[3] + self.imref.gt[5]]],
                'reference extent': {'cols': self.imref.xgsd, 'rows': self.imref.ygsd},  # FIXME not needed anymore
                'success': self.success
            }
            return self.coreg_info

    def correct_shifts(self,
                       max_GCP_count: int = None,
                       cliptoextent: bool = False,
                       min_points_local_corr: int = 5
                       ) -> OrderedDict:
        """Perform a local shift correction using all points from the previously calculated tie point grid.

        NOTE: Only valid matches are used as GCP points.

        :param max_GCP_count: maximum number of GCPs to use
        :param cliptoextent:  whether to clip the output image to its real extent
        :param min_points_local_corr:   number of valid tie points, below which a global shift correction is
                                        performed instead of a local correction (global X/Y shift is then computed as
                                        the mean shift of the remaining points)(default: 5 tie points)
        :return:
        """
        if not self._tiepoint_grid:
            self.calculate_spatial_shifts()

        if self.tiepoint_grid.GCPList:
            if max_GCP_count:
                self.coreg_info['GCPList'] = self.coreg_info['GCPList'][:max_GCP_count]

            # make sure the correction is applied to the original target image
            im2shift = GeoArray(self.params['im_tgt'], nodata=self.nodata[1], progress=self.progress, q=self.q)

            if has_metaRotation(im2shift):
                # resample the target image because (so far) the computed shifts cannot be applied to a dataset with
                # a metadata rotation (GDAL GeoTransform not 0 at positons 2 and 4)
                im2shift = remove_metaRotation(im2shift)

            # apply the correction
            DS = DESHIFTER(im2shift, self.coreg_info,
                           path_out=self.path_out,
                           fmt_out=self.fmt_out,
                           out_crea_options=self.out_creaOpt,
                           align_grids=self.align_grids,
                           match_gsd=self.match_gsd,
                           out_gsd=self.out_gsd,
                           target_xyGrid=self.target_xyGrid,
                           min_points_local_corr=min_points_local_corr,
                           resamp_alg=self.rspAlg_DS,
                           cliptoextent=cliptoextent,
                           # clipextent=self.im2shift.box.boxMapYX,
                           progress=self.progress,
                           v=self.v,
                           q=self.q)

            self.deshift_results = DS.correct_shifts()
            return self.deshift_results
        else:
            if not self.q:
                warnings.warn('Correction of geometric shifts failed because the input GCP list is empty!')
